# -*- coding: utf-8 -*-
# MorDL project: Base tagger model
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a base model for MorDL taggers.
"""
from collections import OrderedDict
from collections.abc import Iterable
from junky import CharEmbeddingRNN, CharEmbeddingCNN, Masking, \
                  get_func_params, to_device
from mordl.base_model import BaseModel
from mordl.defaults import CONFIG_ATTR
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class BatchNorm(nn.BatchNorm1d):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x):
        return super().forward(x.transpose(1, 2)).transpose(1, 2)


class BaseTaggerModel(BaseModel):
    """
    The base model for MorDL taggers.

    Args:

    **num_labels** (`int`): the number of target labels. Don't forget to add
    `1` for padding.

    **labels_pad_idx** (`int`; default=-100): the index of padding element in
    the label vocabulary. You can specify here the real index of the padding
    intent, but we recommend to keep it as is (with default fake index)
    because in practice, learning on padding increasing the resulting model
    performance. If you stil want to experiment, along with specifying the
    real padding index, you may try not to add `1` to **num_labels** for the
    padding intent. The model then put random labels as tags for the padding
    part of the input, but they are ignored during the loss computation.

    **vec_emb_dim** (`int`; default is `None`): the incoming word-level
    embedding vector space dimensionality.

    **alphabet_size** (`int`): the length of character vocabulary for the
    internal character-level embedding layer. Relevant if either
    **rnn_emb_dim** or **cnn_emb_dim** is not `None`.

    **char_pad_idx** (`int`; default is `0`): the index of the padding element
    in the character vocabulary of the internal character-level embedding
    layer. Relevant if either **rnn_emb_dim** or **cnn_emb_dim** is not `None`.

    **rnn_emb_dim** (`int`; default is `None`): the internal character RNN
    (LSTM) embedding dimensionality. If `None`, the layer is skipped.

    **cnn_emb_dim** (`int`; default is `None`): the internal character CNN
    embedding dimensionality. If `None`, the layer is skipped.

    **cnn_kernels** (`list([int])`; default is `[1, 2, 3, 4, 5, 6]`): CNN
    kernel sizes of the internal CNN embedding layer. Relevant if
    **cnn_emb_dim** is not `None`.

    **tag_emb_params** (`dict({'dim': int, 'num': int, 'pad_idx': int})` |
    `list([dict])`; default is `None`): params of internal embedding layers
    for additional `junky.dataset.TokenDataset` outputs. If `None`, the layers
    are not created.

    **emb_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied after the embedding concatenation.

    **emb_do** (`float`; default is `.2`): the dropout rate after the
    embedding concatenation.

    **final_emb_dim** (`int`; default is `512`): the output dimesionality of
    the linear transformation applying to concatenated embeddings.

    **pre_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied before the main part of the algorithm.

    **pre_do** (`float`; default is `.5`): the dropout rate before the main
    part of the algorithm.

    **lstm_layers** (`int`; default is `1`): the number of Bidirectional LSTM
    layers. If `None`, they are not created.

    **lstm_do** (`float`; default is `0`): the dropout between LSTM layers.
    Only relevant, if `lstm_layers` > `1`.

    **tran_layers** (`int`; default is `None`): the number of Transformer Encoder
    layers. If `None`, they are not created.

    **tran_heads** (`int`; default is `8`): the number of attention heads of
    Transformer Encoder layers. Only relevant, if `tran_layers` > `1`.

    **post_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied after the main part of the algorithm.

    **post_do** (`float`; default is `.4`): the dropout rate after the main
    part of the algorithm.
    """
    def __init__(self, num_labels, labels_pad_idx=-100, vec_emb_dim=None,
                 alphabet_size=0, char_pad_idx=0, rnn_emb_dim=None,
                 cnn_emb_dim=None, cnn_kernels=[1, 2, 3, 4, 5, 6],
                 tag_emb_params=None, emb_bn=True, emb_do=.2,
                 final_emb_dim=512, pre_bn=True, pre_do=.5,
                 lstm_layers=1, lstm_do=0, tran_layers=None, tran_heads=8,
                 post_bn=True, post_do=.4):
        if isinstance(cnn_kernels, Iterable):
            cnn_kernels = list(cnn_kernels)
        args, kwargs = get_func_params(BaseTaggerModel.__init__, locals())
        super().__init__(*args, **kwargs)

        assert final_emb_dim % 2 == 0, \
            'ERROR: `final_emb_dim` must be even ' \
           f"(now it's `{final_emb_dim}`)."

        self.num_labels = num_labels
        self.labels_pad_idx = labels_pad_idx
        if vec_emb_dim is None:
            vec_emb_dim = 0
        self.vec_emb_dim = vec_emb_dim

        if rnn_emb_dim:
            self.rnn_emb_l = \
                CharEmbeddingRNN(alphabet_size=alphabet_size,
                                 emb_dim=rnn_emb_dim,
                                 pad_idx=char_pad_idx)
        else:
            self.rnn_emb_l = None
            rnn_emb_dim = 0

        if cnn_emb_dim:
            self.cnn_emb_l = \
                CharEmbeddingCNN(alphabet_size=alphabet_size,
                                 emb_dim=cnn_emb_dim,
                                 pad_idx=char_pad_idx,
                                 kernels=cnn_kernels)
        else:
            self.cnn_emb_l = None
            cnn_emb_dim = 0

        self.tag_emb_l = self.tag_emb_ls = None
        tag_emb_dim = 0
        if tag_emb_params:
            if isinstance(tag_emb_params, dict):
                tag_emb_dim = tag_emb_params['dim'] or 0
                if tag_emb_dim:
                    self.tag_emb_l = \
                        nn.Embedding(tag_emb_params['num'], tag_emb_dim,
                                     padding_idx=tag_emb_params['pad_idx'])
            else:
                self.tag_emb_ls = nn.ModuleList()
                for emb_params in tag_emb_params:
                    tag_emb_dim_ = emb_params['dim']
                    if tag_emb_dim_:
                        tag_emb_dim += tag_emb_dim_
                        self.tag_emb_ls.append(
                            nn.Embedding(emb_params['num'], tag_emb_dim_,
                                         padding_idx=emb_params['pad_idx'])
                        )
                    else:
                        self.tag_emb_ls.append(None)

        joint_emb_dim = vec_emb_dim + rnn_emb_dim + cnn_emb_dim + tag_emb_dim
        assert joint_emb_dim, \
            'ERROR: At least one of `*_emb_dim` must be specified.'

        # PREPROCESS #########################################################
        modules = OrderedDict()
        if emb_bn:
            modules['emb_bn'] = BatchNorm(num_features=joint_emb_dim)
        if emb_do:
            modules['emb_do'] = nn.Dropout(p=emb_do)

        layers = []
        def add_layers(dim, new_dim):
            ls = []
            ls.append(('pre_fc{}_l',
                       nn.Linear(in_features=new_dim, out_features=dim)))
            if pre_bn:
                ls.append(('pre_bn{}', BatchNorm(num_features=dim)))
            ls.append(('pre_nl{}', nn.ReLU()))
            if pre_do:
                ls.append(('pre_do{}', nn.Dropout(p=pre_do)))
            layers.append(ls)

        dim = final_emb_dim
        while joint_emb_dim / dim > 2:
            new_dim = int(dim * 1.5)
            add_layers(dim, new_dim)
            dim = new_dim
        add_layers(dim, joint_emb_dim)
        for idx, layer in enumerate(reversed(layers)):
            for name, module in layer:
                modules[name.format(idx)] = module

        #modules['pre_fc_l'] = nn.Linear(in_features=dim,
        #                                out_features=final_emb_dim)
        #if pre_bn:
        #    modules['pre_bn'] = BatchNorm(num_features=final_emb_dim)
        #modules['pre_nl'] = nn.ReLU()
        #if pre_do:
        #    modules['pre_do'] = nn.Dropout(p=pre_do)
        self.pre_seq_l = nn.Sequential(modules)
        ######################################################################

        if lstm_layers:
            self.lstm_l = nn.LSTM(
                input_size=final_emb_dim,
                hidden_size=final_emb_dim // 2,
                num_layers=lstm_layers, batch_first=True,
                dropout=lstm_do, bidirectional=True
            )
            self.T = nn.Linear(final_emb_dim, final_emb_dim)
            nn.init.constant_(self.T.bias, -1)
        else:
            self.lstm_l = None

        if tran_layers:
            tran_enc_l = nn.TransformerEncoderLayer(
                final_emb_dim, tran_heads, dim_feedforward=2048,
                dropout=0.1, activation='relu'#, layer_norm_eps=1e-5
            )
            tran_norm_l = nn.LayerNorm(normalized_shape=final_emb_dim,
                                       eps=1e-6, elementwise_affine=True)
            self.tran_l = nn.TransformerEncoder(
                tran_enc_l, tran_layers, norm=tran_norm_l
            )
        else:
            self.tran_l = None

        # POSTPROCESS ########################################################
        modules = OrderedDict()
        if post_bn:
            modules['post_bn'] = BatchNorm(num_features=final_emb_dim)
        if post_do:
            modules['post_do'] = nn.Dropout(p=post_do)

        modules['out_fc_l'] = nn.Linear(in_features=final_emb_dim,
                                        out_features=num_labels)
        self.post_seq_l = nn.Sequential(modules)
        ######################################################################

        setattr(self, CONFIG_ATTR, (args, kwargs))

    def forward(self, x, x_lens, x_ch, x_ch_lens, *x_t, labels=None):
        """
        x:         [N (batch), L (sentences), C (words + padding)]
        lens:      number of words in sentences
        x_ch:      [N, L, C (words + padding), S (characters + padding)]
        x_ch_lens: [L, number of characters in words]
        *x_t:      [N, L, C (upos indices)], [N, L, C (xpos indices)], ...
        labels:    [N, L, C]

        Returns logits if label is `None`, (logits, loss) otherwise.
        """
        device = next(self.parameters()).device

        x_ = []
        vec_emb_dim = self.vec_emb_dim
        if vec_emb_dim:
            assert x.shape[2] == vec_emb_dim, \
                   'ERROR: Invalid vector size: ' \
                  f'`{x.shape[2]}` whereas `vec_emb_dim={vec_emb_dim}`'
            x_.append(to_device(x, device))
        if self.rnn_emb_l:
            x_.append(self.rnn_emb_l(to_device(x_ch, device), x_ch_lens))
        if self.cnn_emb_l:
            x_.append(self.cnn_emb_l(to_device(x_ch, device),
                                      to_device(x_ch_lens, device)))
        if self.tag_emb_l:
            x_.append(self.tag_emb_l(to_device(x_t[0], device)))
        elif self.tag_emb_ls:
            for l_, x_t_ in zip(self.tag_emb_ls, to_device(x_t, device)):
                if l_:
                    x_.append(l_(x_t_))

        x = x_[0] if len(x_) == 1 else torch.cat(x_, dim=-1)

        x = self.pre_seq_l(x)

        if self.lstm_l:
            x_ = pack_padded_sequence(x, x_lens.cpu(), batch_first=True,
                                      enforce_sorted=False)
            x_, _ = self.lstm_l(x_)
            x_, _ = pad_packed_sequence(x_, batch_first=True)

            gate = torch.sigmoid(self.T(x))
            x = x_ * gate + x * (1 - gate)

        if self.tran_l:
            src_key_padding_mask = (
                torch.arange(x.shape[1], device=device).expand(x.shape[:-1])
             >= x_lens.to(device).view(1, -1).transpose(0, 1)
                                             .expand(x.shape[:-1])
            )
            x.transpose_(0, 1)  # (N, L, C) to (L, N, C)
            x = self.tran_l(x, src_key_padding_mask=src_key_padding_mask)
            x.transpose_(0, 1)  # (L, N, C) to (N, L, C)

        logits = self.post_seq_l(x)

        if labels is not None:
#             criterion = nn.CrossEntropyLoss().to(device)
#             loss = criterion(logits.flatten(end_dim=-2),
#                              labels.flatten(end_dim=-1))
            loss = F.cross_entropy(logits.view(-1, self.num_labels),
                                   labels.view(-1),
                                   ignore_index=self.labels_pad_idx)
            logits = logits, loss

        return logits



# -*- coding: utf-8 -*-
# MorDL project: NE tagger model
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides FEAT tagger model inherited from `mordl.BaseTaggerModel`.
"""
from collections.abc import Iterable
from junky import get_func_params
from mordl.base_tagger_model import BaseTaggerModel
from mordl.defaults import CONFIG_ATTR


class FeatTaggerModel(BaseTaggerModel):
    """
    The class for prediction the content of a key-value type field. Separated
    implementation (predict only one particular field at a time).

    Args:

    **num_labels** (`int`): the number of target labels. Don't forget to add
    `1` for padding.

    **labels_pad_idx** (`int`; default=-100): the index of padding element in
    the label vocabulary. You can specify here the real index of the padding
    intent, but we recommend to keep it as is (with default fake index)
    because in practice, learning on padding increasing the resulting model
    performance. If you stil want to experiment, along with specifying the
    real padding index, you may try not to add `1` to **num_labels** for the
    padding intent. The model then put random labels as tags for the padding
    part of the input, but they are ignored during the loss computation.

    **vec_emb_dim** (`int`; default is `None`): the incoming word-level
    embedding vector space dimensionality.

    **alphabet_size** (`int`): the length of character vocabulary for the
    internal character-level embedding layer. Relevant if either
    **rnn_emb_dim** or **cnn_emb_dim** is not `None`.

    **char_pad_idx** (`int`; default is `0`): the index of the padding element
    in the character vocabulary of the internal character-level embedding
    layer. Relevant if either **rnn_emb_dim** or **cnn_emb_dim** is not `None`.

    **rnn_emb_dim** (`int`; default is `None`): the internal character RNN
    (LSTM) embedding dimensionality. If `None`, the layer is skipped.

    **cnn_emb_dim** (`int`; default is `200`): the internal character CNN
    embedding dimensionality. If `None`, the layer is skipped.

    **cnn_kernels** (`list([int])`; default is `[1, 2, 3, 4, 5, 6]`): CNN
    kernel sizes of the internal CNN embedding layer. Relevant if
    **cnn_emb_dim** is not `None`.

    **upos_emb_dim** (`int`; default is `200`): the auxiliary UPOS label
    embedding dimensionality.

    **upos_num** (`int`): the length of UPOS vocabulary.

    **upos_pad_idx** (`int`; default is `0`): the index of padding element
    in the UPOS vocabulary.

    **emb_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied after the embedding concatenation.

    **emb_do** (`float`; default is `.2`): the dropout rate after the
    embedding concatenation.

    **final_emb_dim** (`int`; default is `512`): the output dimesionality of
    the linear transformation applying to concatenated embeddings.

    **pre_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied before the main part of the algorithm.

    **pre_do** (`float`; default is `.5`): the dropout rate before the main
    part of the algorithm.

    **lstm_layers** (`int`; default is `1`): the number of Bidirectional LSTM
    layers. If `None`, they are not created.

    **lstm_do** (`float`; default is `0`): the dropout between LSTM layers.
    Only relevant, if `lstm_layers` > `1`.

    **tran_layers** (`int`; default is `None`): the number of Transformer
    Encoder layers. If `None`, they are not created.

    **tran_heads** (`int`; default is `8`): the number of attention heads of
    Transformer Encoder layers. Only relevant, if `tran_layers` > `1`.

    **post_bn** (`bool`; default is `True`): whether batch normalization layer
    should be applied after the main part of the algorithm.

    **post_do** (`float`; default is `.4`): the dropout rate after the main
    part of the algorithm.
    """
    def __init__(self, num_labels, labels_pad_idx=-100, vec_emb_dim=None,
                 alphabet_size=0, char_pad_idx=0, rnn_emb_dim=None,
                 cnn_emb_dim=200, cnn_kernels=[1, 2, 3, 4, 5, 6],
                 upos_emb_dim=200, upos_num=0, upos_pad_idx=0,
                 emb_bn=True, emb_do=.2,
                 final_emb_dim=512, pre_bn=True, pre_do=.5,
                 lstm_layers=1, lstm_do=0, tran_layers=None, tran_heads=8,
                 post_bn=True, post_do=.4):
        if isinstance(cnn_kernels, Iterable):
            cnn_kernels = list(cnn_kernels)
        args, kwargs = get_func_params(FeatTaggerModel.__init__, locals())
        kwargs_ = {x: y for x, y in kwargs.items() if x not in [
            'upos_emb_dim', 'upos_num', 'upos_pad_idx'
        ]}
        if upos_emb_dim:
            kwargs_['tag_emb_params'] = {
                'dim': upos_emb_dim, 'num': upos_num, 'pad_idx': upos_pad_idx
            }
        super().__init__(*args, **kwargs_)
        setattr(self, CONFIG_ATTR, (args, kwargs))



# -*- coding: utf-8 -*-
# MorDL project: Base tagger
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a base class for specialized morphological taggers.
"""
from corpuscula.corpus_utils import _AbstractCorpus
from copy import deepcopy
from distutils.dir_util import copy_tree
import gc
import itertools
import json
import junky
from junky.dataset import CharDataset, DummyDataset, FrameDataset, \
                          LabelDataset, LenDataset, TokenDataset
from morra.base_parser import BaseParser
from mordl import WordEmbeddings
from mordl.defaults import BATCH_SIZE, CONFIG_ATTR, CONFIG_EXT, LOG_FILE, \
                           NONE_TAG, TRAIN_BATCH_SIZE
from mordl.lib.conll18_ud_eval import main as _conll18_ud_eval
#import numpy as np
import os
#from sklearn.metrics import accuracy_score, f1_score, \
#                            precision_score, recall_score
import random
import sys
import time
import torch
from typing import Iterator

_MODEL_CONFIG_FN = 'model_config.json'
_MODEL_FN = 'model.pt'
_DATASETS_CONFIG_FN = 'ds_config.json'
_DATASETS_FN = 'ds.pt'
_CDICT_FN = 'cdict.pickle'


class BaseTagger(BaseParser):
    """
    The base class for the project's taggers.

    Args:

    **embs** (`dict({str: object}); default is `None`): the `dict` with paths
    to embeddings files as keys and corresponding embedding models as values.
    If the tagger needs to load any embedding model, firstly, the model is
    looked up it in that `dict`.

    During init, **embs** is copied to the `embs` attribute of the creating
    object, and this attribute may be used further to share already loaded
    embeddings with another taggers.
    """

    def __err_hideattr(self, name):
        raise AttributeError("'{}' object has no attribute '{}'"
                                 .format(self.__class__.__name__, name))
    def __err_override(self, name):
        raise AttributeError(("If you want to call the method '{}'() "
                              "of class '{}', you must override it")
                                  .format(name, self.__class__.__name__))
    parse_train_corpus = \
        property(lambda self: self.__err_hideattr('parse_train_corpus'))
    backup = \
        property(lambda self: self.__err_hideattr(self, 'backup'))
    restore = \
        property(lambda self: self.__err_hideattr(self, 'restore'))
    _train_init = \
        property(lambda self: self.__err_hideattr(self, '_train_init'))
    _train_eval = \
        property(lambda self: self.__err_hideattr(self, '_train_eval'))
    _train_done = \
        property(lambda self: self.__err_hideattr(self, '_train_done'))

    @property
    def embs(self):
        return self._embs

    def __init__(self, embs=None):
        super().__init__()
        self._model = None
        self._ds = None
        self._embs = {} if embs is None else embs.copy()

    def load_train_corpus(self, corpus, append=False, test=None, seed=None):
        """Loads the train corpus.

        Args:

        **corpus**: either the name of the file in *CoNLL-U* format or the
        `list`/`iterator` of sentences in *Parsed CoNLL-U*.

        **append** (`bool`; default is `False`): whether to add the **corpus**
        to the already loaded one(s).

        **test** (`float`; default is `None`): if not `None`, the **corpus**
        will be shuffled and the specified part of it stored as a test corpus.

        **seed** (`int`; default is `None`): the init value for the random
        number generator if you need reproducibility. Only used if test is not
        `None`.
        """
        args, kwargs = junky.get_func_params(BaseTagger.load_train_corpus,
                                             locals())
        super().load_train_corpus(*args, parse=False, **kwargs)

    @classmethod
    def _get_corpus(cls, corpus, asis=False, log_file=LOG_FILE):
        if isinstance(corpus, str):
            corpus = cls.load_conllu(corpus, log_file=log_file)
        elif (isinstance(corpus, type)
          and issubclass(corpus, _AbstractCorpus)) \
          or isinstance(corpus, _AbstractCorpus):
            corpus = corpus.test()
        elif callable(corpus):
            corpus = corpus()
        return (s[0] if not asis and isinstance(s, tuple) else s
                    for s in corpus)

    @staticmethod
    def _get_filenames(name):
        return tuple(os.path.join(name, x)
                         for x in [_MODEL_CONFIG_FN, _MODEL_FN,
                                   _DATASETS_CONFIG_FN, _DATASETS_FN,
                                   _CDICT_FN])

    @staticmethod
    def _prepare_corpus(corpus, fields, tags_to_remove=None):
        return junky.extract_conllu_fields(
            junky.conllu_remove(corpus, remove=tags_to_remove),
            fields=fields
        )

    def _create_dataset(
        self, sentences, word_emb_type=None, word_emb_path=None,
        word_emb_model_device=None, word_transform_kwargs=None,
        word_next_emb_params=None, with_chars=False, tags=None,
        labels=None, for_self=True, log_file=LOG_FILE
    ):
        ds = FrameDataset()

        if word_emb_type is not None:
            x = WordEmbeddings.create_dataset(
                sentences, emb_type=word_emb_type, emb_path=word_emb_path,
                emb_model_device=word_emb_model_device,
                transform_kwargs=word_transform_kwargs,
                next_emb_params=word_next_emb_params, embs=self.embs,
                loglevel=0 if log_file is None else
                         1 if log_file == sys.stdout else
                         2
            )
            ds.add('x', x)
        else:
            ds.add('x', DummyDataset(data=sentences))
            ds.add('x_lens', LenDataset(data=sentences))

        if with_chars:
            x_ch = CharDataset(sentences,
                               unk_token='<UNK>', pad_token='<PAD>',
                               transform=True)
            ds.add('x_ch', x_ch, with_lens=False)
        else:
            ds.add('x_ch', DummyDataset(data=sentences))
            ds.add('x_ch_lens', DummyDataset(data=sentences))

        if tags:
            for i, t_ in enumerate(tags):
                t = TokenDataset(t_, pad_token='<PAD>', transform=True,
                                 keep_empty=False)
                ds.add('t_{}'.format(i), t, with_lens=False)

        if labels:
            if isinstance(labels[0], tuple):
                y = TokenDataset(labels, pad_token='<PAD>', transform=True,
                                 keep_empty=False)
                ds.add('y', y, with_lens=False)
            else:
                y = LabelDataset(labels, transform=True, keep_empty=False)
                ds.add('y', y)

        if for_self:
            self._ds = ds
        return ds

    def _transform(self, sentences, tags=None, labels=None, ds=None,
                   batch_size=BATCH_SIZE, log_file=LOG_FILE):
        if ds is None:
            ds = self._ds
        for name in ds.list():
            ds_ = ds.get_dataset(name)
            name_ = name.split('_', maxsplit=1)
            typ, idx = name_[0], name_[1] if len(name_) > 1 else None
            if typ == 'x':
                if not WordEmbeddings.transform(
                    ds_, sentences, batch_size=batch_size,
                    loglevel=0 if log_file is None else
                             1 if log_file == sys.stdout else
                             2
                ):
                    ds_.transform(sentences)
            elif typ == 't':
                ds_.transform(tags[int(idx)])
            elif labels and typ == 'y':
                ds_.transform(labels)

    def _transform_collate(self, sentences, tags=None, labels=None, ds=None,
                           batch_size=BATCH_SIZE, log_file=LOG_FILE):
        if ds is None:
            ds = self._ds
        res = []
        for name in self._ds.list():
            ds_, collate_kwargs = self._ds.get(name)
            name_ = name.split('_', maxsplit=1)
            typ, idx = name_[0], name_[1] if len(name_) > 1 else None
            if typ == 'x':
                res_ = WordEmbeddings.transform_collate(
                    ds_, sentences, batch_size=batch_size,
                    loglevel=int(log_file is not None)
                )
                if not res_:
                    res_ = ds_.transform_collate(
                        sentences, batch_size=batch_size,
                        collate_kwargs=collate_kwargs
                    )
                res.append(res_)
            elif typ == 't':
                res.append(ds_.transform_collate(
                    tags[int(idx)], batch_size=batch_size,
                    collate_kwargs=collate_kwargs
                ))
            elif labels and typ == 'y':
                res.append(ds_.transform_collate(
                    labels, batch_size=batch_size,
                    collate_kwargs=collate_kwargs
                ))
        for batch in zip(*res):
            batch_ = []
            for res_ in batch:
                batch_.extend(res_) if isinstance(res_, tuple) else \
                batch_.append(res_)
            yield batch_

    def _save_dataset(self, name, ds=None):
        if not os.path.isdir(name):
            os.mkdir(name)
        if ds is None:
            ds = self._ds
        ds_config_fn, ds_fn = self._get_filenames(name)[2:4]
        config = {}
        for name in ds.list():
            ds_ = ds.get_dataset(name)
            cfg = getattr(ds_, CONFIG_ATTR, None)
            if cfg:
                config[name] = cfg
        with open(ds_config_fn, 'wt', encoding='utf-8') as f:
            print(json.dumps(config, sort_keys=True, indent=4), file=f)
        ds.save(ds_fn, with_data=False)

    def _load_dataset(self, name, emb_path=None, device=None, for_self=True,
                      log_file=LOG_FILE):
        ds_config_fn, ds_fn = self._get_filenames(name)[2:4]
        if log_file:
            print('Loading dataset...', end=' ', file=log_file)
            log_file.flush()
        ds = FrameDataset.load(ds_fn)
        with open(ds_config_fn, 'rt', encoding='utf-8') as f:
            config = json.loads(f.read())
            for name, cfg in config.items():
                WordEmbeddings.apply_config(ds.get_dataset(name), cfg,
                                            emb_path=emb_path, device=device,
                                            embs=self.embs)
        if log_file:
            print('done.', file=log_file)
        if for_self:
            self._ds = ds
        return ds

    def save(self, name, log_file=LOG_FILE):
        """Saves the internal state of the tagger.

        Args:

        **name**: the name to save with.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        The method creates the directory **name** that contains 5 files: two
        for the tagger's model (`model.config.json` and `model.pt`) and two
        for its dataset (`ds.config.json` and `ds.pt`). The 5th file
        (`cdict.pickle`) is an internal state of the
        [`corpuscula.CorpusDict`](https://github.com/fostroll/corpuscula/blob/master/doc/README_CDICT.md)
        object that is used by the tagger as a helper.

        `*.config.json` files contain parameters for creation of the objects.
        They are editable, but you are allowed to change only the device name.
        Any other changes most likely won't allow the tagger to load.
        """
        assert self._ds is not None, \
            "ERROR: The tagger doesn't have a dataset to save"
        assert self._model, "ERROR: The tagger doesn't have a model to save"
        self._save_dataset(name)
        model_config_fn, model_fn, _, _, cdict_fn = \
            self._get_filenames(name)
        self._model.save_config(model_config_fn, log_file=log_file)
        self._model.save_state_dict(model_fn, log_file=log_file)
        self._save_cdict(cdict_fn)

    def load(self, model_class, name, device=None,
             dataset_emb_path=None, dataset_device=None, log_file=LOG_FILE):
        """Loads the tagger's internal state saved by its `.save()` method.

        Args:

        **model_class**: the class of the model used for prediction. Must be
        descendant of the `BaseTaggerModel` class.

        **name** (`str`): the name of the previously saved internal state.

        **device** (`str`; default is `None`): the device for the loaded model
        if you want to override the value from the config.

        **dataset_emb_path** (`str`; default is `None`): the path where the
        dataset's embeddings to load from if you want to override the value
        from the config.

        **dataset_device** (`str`; default is `None`): the device for the
        loaded dataset if you want to override the value from the config.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        self._load_dataset(name, emb_path=dataset_emb_path,
                           device=dataset_device, log_file=log_file)
        model_config_fn, model_fn, _, _, cdict_fn = \
            self._get_filenames(name)
        self._model = model_class.create_from_config(
            model_config_fn, state_dict_f=model_fn,
            device=device, log_file=log_file
        )
        self._load_cdict(cdict_fn, log_file=log_file)

    @staticmethod
    def _normalize_field_names(names):
        res, tostr = [], False
        if names:
            if isinstance(names, str):
                names, tostr = [names], True
            for name in names:
                num_colons = name.count(':')
                if num_colons == 0:
                    name += '::' + NONE_TAG
                elif num_colons == 1:
                    name += ':' + NONE_TAG
                elif num_colons == 2 and name.endswith(':'):
                    name += NONE_TAG
                res.append(name)
        return res[0] if tostr else res

    def _check_cdict(self, sentence, use_cdict_coef):
        return sentence

    def predict(self, field, add_fields, corpus, use_cdict_coef=False,
                with_orig=False, batch_size=BATCH_SIZE, split=None,
                clone_ds=False, save_to=None, log_file=LOG_FILE):
        """Predicts tags for the specified fields of the corpus.

        Args:

        **field** and **add_field** must be the same that were used in the
        `.train()` method.

        **corpus**: the corpus which will be used for the feature extraction
        and predictions. May be either the name of the file in *CoNLL-U*
        format or the `list`/`iterator` of sentences in *Parsed CoNLL-U*.

        **use_cdict_coef** (`bool` | `float`; default is `False`): if `False`,
        we use our prediction only. If `True`, we replace our prediction to
        the value returned by the `corpuscula.CorpusDict.predict_<field>()`
        method if its `coef` >= `.99`. Also, you can specify your own
        threshold as the value of the param.

        **with_orig** (`bool`; default is `False`): if `True`, instead of just
        the sequence with predicted labels, return the sequence of tuples
        where the first element is the sentence with predicted labels and the
        second element is the original sentence. **with_orig** can be `True`
        only if **save_to** is `None`.

        **batch_size** (`int`; default is `64`): the number of sentences per
        batch.

        **split** (`int`; default is `None`): the number of lines in sentences
        split. Allows to process a large dataset in pieces ("splits"). If
        **split** is `None` (default), all the dataset is processed without
        splits.

        **clone_ds** (`bool`; default is `False`): if `True`, the dataset is
        cloned and transformed. If `False`, `transform_collate` is used
        without cloning the dataset. There is no big differences between the
        variants. Both should produce identical results.

        **save_to** (`str`; default is `None`): the file name where the
        predictions will be saved.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        Returns the **corpus** with tag predictions in the specified field.
        """
        assert self._ds is not None, \
               "ERROR: The tagger doesn't have a dataset. Call the train() " \
               'method first'
        assert self._model, \
               "ERROR: The tagger doesn't have a model. Call the train() " \
               'method first'
        assert not with_orig or save_to is None, \
               'ERROR: `with_orig` can be True only if save_to is None'

        field = self._normalize_field_names(field)
        add_fields = self._normalize_field_names(add_fields)

        def process(corpus):

            corpus = self._get_corpus(corpus, asis=True, log_file=log_file)
            device = next(self._model.parameters()).device or junky.CPU

            ds_y = self._ds.get_dataset('y')
            if clone_ds:
                ds = self._ds.clone()
                ds.remove('y')

            for start in itertools.count(step=split if split else 1):
                if isinstance(corpus, Iterator):
                    if split:
                        corpus_ = []
                        for i, sentence in enumerate(corpus, start=1):
                            corpus_.append(sentence)
                            if i == split:
                                break
                    else:
                        corpus_ = list(corpus)
                else:
                    if split:
                        corpus_ = corpus[start:start + split]
                    else:
                        corpus_ = corpus
                if not corpus_:
                    break
                res = \
                    junky.extract_conllu_fields(
                        corpus_, fields=add_fields,
                        with_empty=True, return_nones=True
                    )
                sentences, tags, empties, nones = \
                    res[0], res[1:-2], res[-2], res[-1]
                if clone_ds:
                    self._transform(
                        sentences, tags=tags, batch_size=batch_size, ds=ds,
                        log_file=log_file
                    )
                    loader = ds.create_loader(batch_size=batch_size,
                                              shuffle=False)
                else:
                    loader = self._transform_collate(
                        sentences, tags=tags, batch_size=batch_size,
                        log_file=log_file
                    )
                preds = []
                for batch in loader:
                    #batch = junky.to_device(batch, device)
                    with torch.no_grad():
                        pred = self._model(*batch)
                    pred_indices = pred.argmax(-1)
                    preds.extend(pred_indices.cpu().numpy().tolist())
                values = ds_y.reconstruct(preds)
                if with_orig:
                    res_corpus_ = deepcopy(corpus_)
                    for orig_sentence, sentence in zip(
                        corpus_, junky.embed_conllu_fields(
                            res_corpus_, field, values,
                            empties=empties, nones=nones
                        )
                    ):
                        yield self._check_cdict(sentence, use_cdict_coef), \
                              orig_sentence
                else:
                    for sentence in junky.embed_conllu_fields(
                        corpus_, field, values,
                        empties=empties, nones=nones
                    ):
                        yield self._check_cdict(sentence, use_cdict_coef)

        corpus = process(corpus)
        if save_to:
            self.save_conllu(corpus, save_to, log_file=None)
            corpus = self._get_corpus(save_to, asis=True, log_file=log_file)
        return corpus

    def evaluate(self, field, gold, test=None, feats=None, label=None,
                 use_cdict_coef=False, batch_size=BATCH_SIZE, split=None,
                 clone_ds=False, log_file=LOG_FILE, **ext_predict_kwargs):
        """Evaluate the tagger model.

        Args:

        **field** must be the same that was used in the `.train()` method.

        **gold**: the corpus of sentences with actual target values to score
        the tagger on. May be either the name of the file in *CoNLL-U* format
        or the `list`/`iterator` of sentences in *Parsed CoNLL-U*.

        **test** (default is `None`): the corpus of sentences with predicted
        target values. If `None` (default), the **gold** corpus will be
        retagged on-the-fly, and the result will be used as the **test**.

        **feats** (`str | list([str])`; default is `None`): one or several
        subfields of the key-value type fields like `FEATS` or `MISC` to be
        evaluated separatedly.

        **label** (`str`; default is `None`): the specific label of the target
        field to be evaluated separatedly, e.g. `field='UPOS', label='VERB'`
        or `field='FEATS:Animacy', label='Inan'`.

        **use_cdict_coef** (`bool` | `float`; default is `False`): if `False`,
        we use our prediction only. If `True`, we replace our prediction to
        the value returned by the `corpuscula.CorpusDict.predict_<field>()`
        method if its `coef` >= `.99`. Also, you can specify your own
        threshold as the value of the param.

        **batch_size** (`int`; default is `64`): the number of sentences per
        batch.

        **split** (`int`; default is `None`): the number of lines in sentences
        split. Allows to process a large dataset in pieces ("splits"). If
        **split** is `None` (default), all the dataset is processed without
        splits.

        **clone_ds** (`bool`; default is `False`): if `True`, the dataset is
        cloned and transformed. If `False`, `transform_collate` is used
        without cloning the dataset. There is no big differences between the
        variants. Both should produce identical results.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        **\*\*ext_predict_kwargs**: extended keyword arguments for the
        `.predict()` method. Will be passed as is.

        The method prints metrics and returns evaluation accuracy.
        """
        if isinstance(feats, str):
            feats = [feats]
        gold = self._get_corpus(gold, log_file=log_file)
        corpora = zip(self._get_corpus(test, log_file=log_file), gold) \
                      if test else \
                  self.predict(gold, use_cdict_coef=use_cdict_coef,
                               with_orig=True, batch_size=batch_size,
                               split=split, clone_ds=clone_ds,
                               log_file=log_file, **ext_predict_kwargs)
        self._normalize_field_names(field)
        header = field.split(':')[:2]
        if len(header) == 2 and not header[1]:
            header = header[:1]
        field = header[0]
        name = header[1] if len(header) == 2 else None
        header = ':'.join(header)
        if label:
            header += '==' + label
        if log_file:
            print('Evaluating ' + header, file=log_file)

        def compare(gold_label, test_label, n, c, nt, ct, ca, ce, cr):
            n += 1
            if (label and (gold_label == label
                        or test_label == label)) \
            or (not label and (gold_label or test_label)):
                nt += 1
                if gold_label == test_label:
                    c += 1
                    ct += 1
                elif not gold_label or (label
                                    and gold_label != label):
                    ce += 1
                elif not test_label or (label
                                    and test_label != label):
                    ca += 1
                else:
                    cr += 1
            else:
                c += 1
            return n, c, nt, ct, ca, ce, cr

        # n - the overall number of tags (outer join of gold and predicted)
        #     plus 1 for each void token in case of non-feats field
        # c - the number of tags predicted correctly (plus 1 for each
        #     correctly predicted void token in case of non-feats field)
        # nt - the overall number of tags (outer join of gold and predicted)
        # ct - the number of tags predicted correctly
        # ca - the number of absent feats (not predicted)
        # ce - the number of excess feats (predicted when they aren't in gold)
        # cr - the number of feats of wrong type in prediction
        # ntok - the overall number of tokens
        # ctok - the number of tokens predicted correctly
        # ntokt - the number of non-void tokens (gold or predicted)
        # ctokt - tne number of non-void tokens predicted correctly
        n = c = nt = ct = ca = ce = cr = ntok = ctok = ntokt = ctokt = 0
        # nsent - the overall number of sentences
        # csent - the number of sentences predicted correctly
        nsent = csent = 0
        # labels (for f1 counting)
        res_golds, res_preds = [], []
        for nsent, sentences in enumerate(corpora, start=1):
            csent_ = 1
            for gold_token, test_token in zip(*sentences):
                wform = gold_token['FORM']
                if wform and '-' not in gold_token['ID']:
                    gold_label = gold_token[field]
                    test_label = test_token[field]
                    if name:
                        gold_label = gold_label.get(name)
                        test_label = test_label.get(name)
                    isgold = isinstance(gold_label, dict)
                    istest = isinstance(test_label, dict)
                    if isgold and istest:
                        assert not label, \
                            'ERROR: To evaluate exact label of dict field, ' \
                            "add feat name to field param as '<field:feat>'"
                        ntok += 1
                        if gold_label or test_label:
                            ctok_ = 1
                            # sub labels (for f1 counting)
                            res_gold, res_pred = {}, {}
                            for feat in feats if feats else set(
                                [*gold_label.keys(), *test_label.keys()]
                            ):
                                gold_feat = gold_label.get(feat)
                                test_feat = test_label.get(feat)
                                res_gold[feat] = gold_feat
                                res_pred[feat] = test_feat
                                n, c_, nt, ct, ca, ce, cr = \
                                    compare(gold_feat, test_feat,
                                            n, c, nt, ct, ca, ce, cr)
                                if c_ == c:
                                    ctok_ = 0
                                    csent_ = 0 
                                else:
                                    c = c_
                            res_golds.append(res_gold)
                            res_preds.append(res_pred)
                            ntokt += 1
                            ctok += ctok_
                            ctokt += ctok_
                        else:
                            ctok += 1
                    elif not (isgold or istest):
                        res_golds.append(gold_label)
                        res_preds.append(test_label)
                        n, c_, nt, ct, ca, ce, cr = \
                            compare(gold_label, test_label,
                                    n, c, nt, ct, ca, ce, cr)
                        if c_ == c:
                            csent_ = 0 
                        else:
                            c = c_
                    else:
                        raise TypeError(
                            'Inconsistent field types in gold and test '
                            'corpora'
                        )
            csent += csent_
        if test:
            try:
                next(gold)
            except StopIteration:
                pass

        '''
        if res_golds and isinstance(res_golds[0], dict):
            feats = set(x for x in res_golds for x in x)
            res_golds = [[f'{y}:{x.get(y) or "_"}' for y in feats]
                             for x in res_golds]
            res_preds = [[f'{y}:{x.get(y) or "_"}' for y in feats]
                             for x in res_preds]
            g, p = list(zip(*res_golds)), list(zip(*res_preds))
            g_, p_ = list(zip(*[list(zip(*x)) if x else [(), ()] for x in (
                [(x, y) for x, y in x if x != y or not x.endswith(':_')]
                    for x in (list(zip(x, y)) for x, y in zip(g, p))
            )]))

            accuracy = np.mean([accuracy_score(x, y) for x, y in zip(g, p)])
            precision = np.mean([precision_score(x, y, average='macro')
                                     for x, y in zip(g, p)])
            recall = np.mean([recall_score(x, y, average='macro')
                                  for x, y in zip(g, p)])
            f1 = np.mean([f1_score(x, y, average='macro')
                              for x, y in zip(g, p)])

            accuracy = (accuracy, np.mean([accuracy_score(x, y)
                                               for x, y in zip(g_, p_)]))
            precision = (precision,
                         np.mean([precision_score(x, y, average='macro')
                                      for x, y in zip(g_, p_)]))
            recall = (recall, np.mean([recall_score(x, y, average='macro')
                                           for x, y in zip(g_, p_)]))
            f1 = (f1, np.mean([f1_score(x, y, average='macro')
                                   for x, y in zip(g_, p_)]))

            g__, p__ = [x for x in g_ for x in x], [x for x in p_ for x in x]
            accuracy = (*accuracy, accuracy_score(g__, p__))
            precision = (*precision,
                         precision_score(g__, p__, average='macro'))
            recall = (*recall, recall_score(g__, p__, average='macro'))
            f1 = (*f1, f1_score(g__, p__, average='macro'))
        else:
            cats = {y: x for x, y in enumerate(set((*res_golds, *res_preds)))}
            res_golds = [cats[x] for x in res_golds]
            res_preds = [cats[x] for x in res_preds]

            accuracy = accuracy_score(res_golds, res_preds)
            precision = precision_score(res_golds, res_preds, average='macro')
            recall = recall_score(res_golds, res_preds, average='macro')
            f1 = f1_score(res_golds, res_preds, average='macro')

        print('----------------------------------------')
        print('accuracy:', accuracy)
        print('precision:', precision)
        print('recall:', recall)
        print('f1', f1)
        print('----------------------------------------')
        '''

        if log_file:
            if nsent <= 0:
                print('Nothing to do!', file=log_file)
            else:
                sp = ' ' * (len(header) - 2)
                print(header + ' total: '
                    + ('{} tokens, {} tags'.format(ntok, nt) if ntok else
                       '{}'.format(nt)), file=log_file)
                print(sp + ' correct: '
                    + ('{} tokens, {} tags'.format(ctok, ct) if ntok else
                       '{}'.format(ct)), file=log_file)
                print(sp + '   wrong: '
                    + ('{} tokens, {} tags'.format(ntok - ctok, nt - ct)
                           if ntok else
                      '{}'.format(nt - ct))
                    + (' [{} excess / {} absent{}]'
                           .format(ce, ca, '' if label else
                                           ' / {} wrong type'.format(cr))
                           if cr != nt - ct else
                       ''), file=log_file)
                print(sp + 'Accuracy: {}{}'
                               .format('{} / '.format(ctokt / ntokt
                                                          if ntokt > 0 else
                                                      1.)
                                           if ntok else
                                       '',
                                       ct / nt if nt > 0 else 1.),
                      file=log_file)
                if nt != n or ntokt != ntok:
                    print('[Total accuracy: {}{}]'
                              .format('{} / '.format(ctok / ntok
                                                         if ntok > 0 else
                                                     1.)
                                          if ntok else
                                      '',
                                      c / n if n > 0 else 1.), file=log_file)
                print('[By sentence accuracy: {}]'
                          .format('{}'.format(csent / nsent)), file=log_file)
        return ct / nt if nt > 0 else 1.

    def train(self, field, add_fields, model_class, tag_emb_names, save_as,
              device=None, control_metric='accuracy', max_epochs=None,
              min_epochs=0, bad_epochs=5, batch_size=TRAIN_BATCH_SIZE,
              max_grad_norm=None, tags_to_remove=None, word_emb_type='bert',
              word_emb_path='xlm-roberta-base', word_transform_kwargs=None,
                  # BertDataset.transform() (for BERT-descendant models)
                  # params:
                  # {'max_len': 0, 'batch_size': 64, 'hidden_ids': '10',
                  #  'aggregate_hiddens_op': 'cat',
                  #  'aggregate_subtokens_op': 'absmax', 'to': junky.CPU,
                  #  'loglevel': 1}
                  # WordDataset.transform() (for other models) params:
                  # {'check_lower': True}
              stage1_params=None,
                  # {'lr': .0001, 'betas': (0.9, 0.999), 'eps': 1e-8,
                  #  'weight_decay': 0, 'amsgrad': False,
                  #  'max_epochs': None, 'min_epochs': None,
                  #  'bad_epochs': None, 'batch_size': None,
                  #  'max_grad_norm': None}
              stage2_params=None,
                  # {'lr': .001, 'momentum': .9, 'weight_decay': 0,
                  #  'dampening': 0, 'nesterov': False,
                  #  'max_epochs': None, 'min_epochs': None,
                  #  'bad_epochs': None, 'batch_size': None,
                  #  'max_grad_norm': None}
              stage3_params={'save_as': None},
                  # {'save_as': None, 'epochs': 3, 'batch_size': 8,
                  #  'lr': 2e-5, 'betas': (0.9, 0.999), 'eps': 1e-8,
                  #  'weight_decay': .01, 'amsgrad': False,
                  #  'num_warmup_steps': 3, 'max_grad_norm': 1.}
              stages=[1, 2, 3, 1, 2], save_stages=False, load_from=None,
              learn_on_padding=True, remove_padding_intent=False,
              seed=None, start_time=None, keep_embs=False, log_file=LOG_FILE,
              **model_kwargs):
        """Creates and trains the tagger model.

        We assume all positional argumets but **save_as** are for internal use
        only and should be hide in descendant classes.

        During training, the best model is saved after each successful epoch.

        *Training's args*:

        **field** (`str`): the name of the field which needs to be predicted
        by the training tagger. May contain up to 3 elements, separated by the
        colon (`:`). Format is:
        `'<field name>:<feat name>:<replacement for None>'`. The replacement
        is used during the training time as a filler for the fields without a
        value for that we could predict them, too. In the *CoNLL-U* format,
        the replacement is the `'_'` sign, so we use it as the default
        replacement. You hardly have a reason to change it. Examples:<br/>
        `'UPOS'` - predict the *UPOS* field;<br/>
        `'FEATS:Animacy'` - predict only the *Animacy* feat of the *FEATS*
        field;<br/>
        `'FEATS:Animacy:_O'` - likewise the above, but if the feat value is
        `None`, it will be replaced by `'_O'` during training;<br/>
        `'XPOS::_O'` - predict the *XPOS* field and use `'_O'` as the
        replacement for `None`.

        **add_fields** (`None` | `str` | `list([str])`): any auxiliary fields
        to use along with the *FORM* field for the prediction. If `None`, only
        the *FORM* field is used. To use additional fields, pass the field
        name (e.g. `'UPOS'`) or the `list` of field names (e.g. `['UPOS',
        'LEMMA']`). These fields are included in the dataset. The format of
        each element of the **add_fields** is equal to the **field** format.

        **model_class**: the class of the model using for the prediction. Must
        be a descendant of the `BaseTaggerModel` class.

        **tag_emb_names** (`str|list([str])`): prefixes of the model args,
        using instead of the **tag_emb_params** `dict` in the
        `BaseTaggerModel` class. Each name refers to the corresponding field
        in the **add_fields** arg. You have to look into sources of the
        descendant classess if you really want to make sense of it.

        **save_as** (`str`): the name using for save the model's head. Refer
        to the `.save()` method's help for the broad definition (see the
        **name** arg there).

        **device** (`str`, default is `None`): the device for the model. E.g.:
        'cuda:0'. If `None`, we don't move the model to any device (it is
        placed right where it's created).

        **control_metric** (`str`; default is `accuracy`): the metric that
        control training. Any that is supported by the `junky.train()` method.
        In the moment, it is: 'accuracy', 'f1', 'loss', 'precision', and
        'recall'.

        **max_epochs** (`int`; default is `None`): the maximal number of
        epochs for the model's head training (stages types `1` and `2`). If
        `None` (default), the training would be linger until **bad_epochs**
        has met, but no less than **min_epochs**.

        **min_epochs** (`int`; default is `0`): the minimal number of training
        epochs for the model's head training (stages types `1` and `2`).

        **bad_epochs** (`int`; default is `5`): the maximal allowed number of
        bad epochs (epochs when chosen **control_metric** is not became
        better) in a row for the model's head training (stages types `1` and
        `2`).

        **batch_size** (`int`; default is `32`): the number of sentences per
        batch for the model's head training (stages types `1` and `2`).

        **max_grad_norm** (`float`; default is `None`): the gradient clipping
        parameter for the model's head training (stages types `1` and `2`).

        **tags_to_remove** (`dict({str: str}) | dict({str: list([str])})`;
        default is `None`): the tags, tokens with those must be removed from
        the corpus. It's the `dict` with field names as keys and values you
        want to remove. Applied only to fields with atomic values (like
        *UPOS*). This argument may be used, for example, to remove some
        infrequent or just excess tags from the corpus. Note, that we remove
        the tokens from the train corpus completely, not just replace those
        tags to `None`.

        *Word embedding params*:

        **word_emb_type** (`str`; default is `'bert'`): one of (`'bert'` |
        `'glove'` | `'ft'` | `'w2v'`) embedding types.

        **word_emb_path** (`str`): the path to the word embeddings storage.

        **word_transform_kwargs** (`dict`; default is `None`): keyword
        arguments for the `.transform()` method of the dataset created for
        sentences to word embeddings conversion. See the `.transform()` method
        of either `junky.datasets.BertDataset` (if **word_emb_path** is
        `'bert'`) or `junky.datasets.WordDataset` (otherwise) if you want to
        learn allowed values for the parameter. If `None`, the `.transform()`
        method use its defaults.

        *Training stages params*:

        **stage1_param** (`dict`; default is `None`): keyword arguments for
        the `BaseModel.adjust_model_for_train()` method. If `None`, the
        defaults are used. Also, you can specify here new values for the
        arguments **max_epochs**, **min_epochs**, **bad_epochs**,
        **batch_size**, and **max_grad_norm** that will be used only on stages
        of type `1`.

        **stage2_param** (`dict`; default is `None`): keyword arguments for
        the `BaseModel.adjust_model_for_tune()` method. If `None`, the
        defaults are used. Also, you can specify here new values for the
        arguments **max_epochs**, **min_epochs**, **bad_epochs**,
        **batch_size**, and **max_grad_norm** that will be used only on stages
        of type `2`.

        **stage3_param** (`dict`; default is `None`): keyword arguments for
        the `WordEmbeddings.full_tune()` method. If `None`, the defaults are
        used.

        **stages** (`list([int]`; default is `[1, 2, 3, 1, 2]`): what stages
        we should use during training and in which order. On the stage type
        `1` the model head is trained with *Adam* optimizer; the stage type
        `2` is similar, but the optimizer is *SGD*; the stage type `3` is only
        relevant when **word_emb_type** is `'bert'` and we want to tune the
        whole model. Stage type `0` defines the skip-stage, i.e. there would
        be no real training on it. It is used when you need reproducibility
        and want to continue train the model from some particular stage. In
        this case, you specify the name of the model saved on that stage in
        the parametere **load_from**, and put zeros into the **stages** list
        on the places of already finished ones. One more time: it is used for
        reproducibility only, i.e. when you put some particular value to the
        **seed** param and want the data order in bathes be equivalent with
        data on the stages from the past trainings.

        **save_stages** (`bool`; default is `False`): if we need to keep the
        best model of each stage beside of the overall best model. The names
        of these models would have the suffix `_<idx>(stage<stage_type>)`
        where `<idx>` is an ordinal number of the stage. We can then use it to
        continue training from any particular stage number (changing next
        stages or their parameters) using the parameter **load_from**. Note
        that we save only stages of the head model. The embedding model as a
        part of the full model usually tune only once, so we don't make its
        copy.

        **load_from** (`str`; default is `None`): if you want to continue
        training from one of previously saved stages, you can specify the name
        of the model from that stage. Note, that if your model is already
        trained on stage type `3`, then you want to set param
        **word_emb_path** to `None`. Otherwise, you'll load wrong embedding
        model. Any other params of the model may be overwritten (and most
        likely, this would cause error), but they are equivalent when the
        training is just starts and when it's continues. But the
        **word_emb_path** is different if you already passed stage type `3`,
        so don't forget to set it to `None` in that case. (Example: you want
        to repeat training on stage no `5`, so you specify in the
        **load_from** param something like `'model_4(stage1)'` and set the
        **word_emb_path** to `None` and the **stages_param** to
        `'[0, 0, 0, 0, 2]'` (or, if you don't care of reproducibility, you
        could just specify `[2]` here).

        *Other options*:

        **learn_on_padding** (`bool`; default is `True`): while training, we
        can calculate loss either taking in account predictions made for
        padding tokens or without it. The common practice is don't use padding
        when calculate loss. However, we note that using padding sometimes
        makes the resulting model performance slightly better.

        **remove_padding_intent** (`bool`; default is `False`): if you set
        **learn_on_padding** param to `False`, you may want not to use padding
        intent during training at all. I.e. padding tokens would be tagged
        with some of real tags, and they would just ignored during computing
        loss. As a result, the model would have the output dimensionality of
        the final layer less by one. On the first sight, such approach could
        increase the performance, but in our experiments, such effect appeared
        not always.

        **seed** (`int`; default is `None`): init value for the random number
        generator if you need reproducibility. Note that each stage will have
        its own seed value, and the **seed** param is used to calculate these
        values.

        **start_time** (`float`; default is `None`): the result of
        `time.time()` to start with. If `None`, the arg will be init anew.

        **keep_embs** (`bool`; default is `False`): by default, after creating
        `Dataset` objects, we remove word embedding models to free memory.
        With `keep_embs=False` this operation is omitted, and you can use
        `.embs` attribute for share embedding models with other objects.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        *The model hyperparameters*:

        **\*\*model_kwargs**: keyword arguments for the model creating. Will
        be passed as is to the **model_class** constructor.

        The method returns the train statistics.
        """
        assert self._train_corpus, 'ERROR: Train corpus is not loaded yet'
        assert not (learn_on_padding and remove_padding_intent), \
            'ERROR: Either `learn_on_padding` or `remove_padding_intent` ' \
            'must be `False`.'

        if seed:
            junky.enforce_reproducibility(seed=seed)

        if not start_time:
            start_time = time.time()

        if isinstance(tag_emb_names, str):
            tag_emb_names = [tag_emb_names]

        field = self._normalize_field_names(field)
        header = field.split(':')[:2]
        if len(header) == 2 and not header[1]:
            header = header[:1]
        header = ':'.join(header)
        bert_header = header.lower().replace(':', '-') + '_'
        fields = self._normalize_field_names(
            [] if add_fields is None else \
            [add_fields] if isinstance(add_fields, str) else \
            add_fields
        )
        fields.append(field)

        # Train the model head with Adam
        def stage1(load_from, save_to, res, save_to2=None, seed=None):
            if log_file:
                print(f'\nMODEL TRAINING {idx} (STAGE 1, SEED {seed})',
                      file=log_file)
            model_config_fn, model_fn, _, _, cdict_fn = \
                self._get_filenames(save_to)

            if seed:
                junky.enforce_reproducibility(seed=seed)

            stage1_params_ = deepcopy(stage1_params) or {}
            train_params = {}
            for param, value in zip(['max_epochs', 'min_epochs', 'bad_epochs',
                                     'batch_size', 'max_grad_norm'],
                                    [max_epochs, min_epochs, bad_epochs,
                                     batch_size, max_grad_norm]):
                value_ = stage1_params_.pop(param, None) \
                             if stage1_params_ else \
                         None
                train_params['epochs' if param == 'max_epochs' else param] = \
                    value_ if value_ is not None else value
            ds_ = ds_train.get_dataset('y')
            if hasattr(ds_, 'pad') and not learn_on_padding:
                stage1_params_['labels_pad_idx'] = ds_.pad

            if load_from:
                _, model_fn_, _, _, _ = \
                   self._get_filenames(load_from)
                model.load_state_dict(model_fn_, log_file=log_file)
            criterion, optimizer, scheduler = \
                model.adjust_model_for_train(**(stage1_params_
                                                    if stage1_params_ else
                                                {}))
            best_epoch, best_score = (res['best_epoch'], res['best_score']) \
                                         if res else \
                                     (0, None)

            def best_model_backup_method(model, model_score):
                if log_file:
                    print('new maximum score {:.8f}'.format(model_score),
                          file=log_file)
                self._save_dataset(save_to, ds=ds_train)
                self._save_cdict(cdict_fn)
                model.save_config(model_config_fn, log_file=log_file)
                model.save_state_dict(model_fn, log_file=log_file)

            change_load_from = False
            res_ = junky.train(
                None, model, criterion, optimizer, scheduler,
                best_model_backup_method, datasets=(ds_train, ds_test),
                **train_params, control_metric=control_metric,
                batch_to_device=False, best_score=best_score,
                with_progress=log_file is not None, log_file=log_file
            )
            if res_ and res_['best_epoch'] is not None:
                change_load_from = True
                if save_to2 and save_to2 != save_to:
                    copy_tree(save_to, save_to2)
                if res:
                    for key, value in res_.items():
                        if key == 'best_epoch':
                            res[key] += value
                        elif key == 'best_score':
                            res[key] = value
                        else:
                            res[key][:best_epoch] = value
                else:
                    res = res_
            return res, change_load_from

        # Train the model head with SGD
        def stage2(load_from, save_to, res, save_to2=None, seed=None):
            if log_file:
                print(f'\nMODEL TRAINING {idx} (STAGE 2, SEED {seed})',
                      file=log_file)
            model_config_fn, model_fn, _, _, cdict_fn = \
                self._get_filenames(save_to)

            if seed:
                junky.enforce_reproducibility(seed=seed)

            stage2_params_ = deepcopy(stage1_params) or {}
            train_params = {}
            for param, value in zip(['max_epochs', 'min_epochs', 'bad_epochs',
                                     'batch_size', 'max_grad_norm'],
                                    [max_epochs, min_epochs, bad_epochs,
                                     batch_size, max_grad_norm]):
                value_ = stage2_params_.pop(param, None) \
                             if stage2_params_ else \
                         None
                train_params['epochs' if param == 'max_epochs' else param] = \
                    value_ if value_ is not None else value
            ds_ = ds_train.get_dataset('y')
            if hasattr(ds_, 'pad') and not learn_on_padding:
                stage2_params_['labels_pad_idx'] = ds_.pad

            if load_from:
                _, model_fn_, _, _, _ = \
                   self._get_filenames(load_from)
                model.load_state_dict(model_fn_, log_file=log_file)
            criterion, optimizer, scheduler = \
                model.adjust_model_for_tune(**(stage2_params_
                                                   if stage2_params_ else
                                               {}))
            best_epoch, best_score = (res['best_epoch'], res['best_score']) \
                                         if res else \
                                     (0, None)

            def best_model_backup_method(model, model_score):
                if log_file:
                    print('new maximum score {:.8f}'.format(model_score),
                          file=log_file)
                self._save_dataset(save_to, ds=ds_train)
                self._save_cdict(cdict_fn)
                model.save_config(model_config_fn, log_file=log_file)
                model.save_state_dict(model_fn, log_file=log_file)

            change_load_from = False
            res_ = junky.train(
                None, model, criterion, optimizer, scheduler,
                best_model_backup_method, datasets=(ds_train, ds_test),
                **train_params, control_metric=control_metric,
                batch_to_device=False, best_score=best_score,
                with_progress=log_file is not None, log_file=log_file
            )
            if res_ and res_['best_epoch'] is not None:
                change_load_from = True
                if save_to2 and save_to2 != save_to:
                    copy_tree(save_to, save_to2)
                if res:
                    for key, value in res_.items():
                        if key == 'best_epoch':
                            res[key] += value
                        elif key == 'best_score':
                            res[key] = value
                        else:
                            res[key][:best_epoch] = value
                else:
                    res = res_
            return res, change_load_from

        # Train the full model with AdamW
        def stage3(load_from, save_to, res, save_to2=None, seed=seed):
            if log_file:
                print(f'\nMODEL TRAINING {idx} (STAGE 3, SEED {seed})',
                      file=log_file)
            model_config_fn, model_fn, _, _, cdict_fn = \
                self._get_filenames(save_to)

            if seed:
                junky.enforce_reproducibility(seed=seed)

            if load_from:
                _, model_fn_, _, _, _ = \
                   self._get_filenames(load_from)
                model.load_state_dict(model_fn_, log_file=log_file)
            best_epoch, best_score = (res['best_epoch'], res['best_score']) \
                                         if res else \
                                     (0, None)

            def tune_word_emb(emb_type, best_score=None):
                res = None
                if emb_type == 'bert':
                    params = deepcopy(stage3_params) or {}
                    if 'save_as' not in params:
                        params['save_as'] = bert_header

                    def model_save_method(_):
                        config = getattr(self._ds.get_dataset('x'),
                                         CONFIG_ATTR)
                        config['emb_path'] = params['save_as']
                        self._save_dataset(save_to)
                        self._save_cdict(cdict_fn)
                        model.save_config(model_config_fn, log_file=log_file)
                        model.save_state_dict(model_fn, log_file=log_file)

                    res = WordEmbeddings.full_tune(
                        model, save_to, model_save_method,
                        (ds_train, ds_test),
                        (train[0], test[0]) if test else train[0],
                        control_metric=control_metric, best_score=best_score,
                        **params, transform_kwargs=word_transform_kwargs,
                        seed=None, log_file=log_file
                            # we've already initialized seed earlier
                    )
                else:
                    raise ValueError(f"ERROR: Tune method for '{emb_type}' "
                                      'embeddings is not implemented')
                return res

            change_load_from = False
            res_ = tune_word_emb(word_emb_type, best_score=best_score)
            if res_ and res_['best_epoch'] is not None:
                change_load_from = True
                if save_to2 and save_to2 != save_to:
                    copy_tree(save_to, save_to2)
                if res:
                    for key, value in res_.items():
                        if key == 'best_epoch':
                            res[key] += value
                        elif key == 'best_score':
                            res[key] = value
                        else:
                            res[key][:best_epoch] = value
                else:
                    res = res_
            return res, change_load_from

        stage_methods = [stage1, stage2, stage3]
        stage_ids = list(range(1, len(stage_methods) + 1))
        for stage in stages:
            assert not stage or stage in stage_ids, \
               f'ERROR: The stage {stage} is invalid. ' \
               f'Only stages {stage_ids} are allowed.'
        assert word_emb_type.lower() == 'bert' or 3 not in stages, \
            'ERROR: The stage 3 is only allowed with `word_emb_type=bert`.'
        if stages.count(3) > 1:
            print('WARNING: Save of the BERT model will not be staged.',
                  file=sys.stderr)

        if log_file:
            print(f'\n=== {header} TAGGER TRAINING PIPELINE ===',
                  file=log_file)

        # 1. Prepare corpora
        train = self._prepare_corpus(
            self._train_corpus, fields=fields,
            tags_to_remove=tags_to_remove
        )
        test = self._prepare_corpus(
            self._test_corpus, fields=fields,
            tags_to_remove=tags_to_remove
        ) if self._test_corpus is not None else None

        if device:
            torch.cuda.set_device(device)

        # 2. Create datasets
        def stage_ds(idx):
            if log_file:
                print('\nDATASETS CREATION', file=log_file)
            log_file_ = sys.stderr if log_file else None
            if self._ds is None:
                ds_train = self._create_dataset(
                    train[0],
                    word_emb_type=word_emb_type, word_emb_path=word_emb_path,
                    word_emb_model_device=device,
                    word_transform_kwargs=word_transform_kwargs,
                    #word_next_emb_params=word_next_emb_params,
                    with_chars=model_kwargs.get('rnn_emb_dim') \
                            or model_kwargs.get('cnn_emb_dim'),
                    tags=train[1:-1], labels=train[-1], for_self=False,
                    log_file=log_file_
                )
                self._ds = ds_train.clone(with_data=False)
            else:
                ds_train = self._ds.clone(with_data=False)
                self._transform(train[0], tags=train[1:-1], labels=train[-1],
                                ds=ds_train, log_file=log_file_)
            if test:
                ds_test = ds_train.clone(with_data=False)
                self._transform(test[0], tags=test[1:-1], labels=test[-1],
                                ds=ds_test, log_file=log_file_)
            else:
                ds_test = None

            # remove emb models to free memory:
            if not keep_embs and 3 not in stages[idx:]:
                self._ds._pull_xtrn()
                ds_train._pull_xtrn()
                if ds_test is not None:
                    ds_test._pull_xtrn()
                self._embs = {}
                gc.collect()

            return ds_train, ds_test

        # 3. Create model
        res = None
        if load_from:
            if log_file:
                print('\nMODEL LOADING', file=log_file)
            self.load(load_from, device=device,
                      dataset_emb_path=word_emb_path, dataset_device=device,
                      log_file=log_file)
            model = self._model

            ds_train, ds_test = stage_ds(1)

            if ds_test:
                res = junky.train(
                    None, model, None, None, None,
                    None, datasets=(None, ds_test),
                    epochs=1, min_epochs=0, bad_epochs=1,
                    batch_size=batch_size, control_metric=control_metric,
                    max_grad_norm=None, batch_to_device=False,
                    best_score=None, with_progress=False, log_file=None
                )
                print(f"\nBest score: {res['best_score']}")

        else:
            ds_train, ds_test = stage_ds(1)

            if log_file:
                print('\nMODEL CREATION', file=log_file)
            if seed:
                junky.enforce_reproducibility(seed=seed)

            ds_ = ds_train.get_dataset('y')
            num_labels = len(ds_.transform_dict)
            model_args = [num_labels - 1 if remove_padding_intent else 
                          num_labels]
            if hasattr(ds_, 'pad') and not learn_on_padding:
                model_kwargs['labels_pad_idx'] = ds_.pad
            if word_emb_type:
                ds_ = ds_train.get_dataset('x')
                model_kwargs['vec_emb_dim'] = ds_.vec_size
            if model_kwargs.get('rnn_emb_dim') \
            or model_kwargs.get('cnn_emb_dim'):
                ds_ = ds_train.get_dataset('x_ch')
                model_kwargs['alphabet_size'] = len(ds_.transform_dict)
                model_kwargs['char_pad_idx'] = ds_.pad
            if tag_emb_names:
                ds_list = ds_train.list()
                names = iter(tag_emb_names)
                for name in ds_train.list():
                    if name.startswith('t_'):
                        ds_ = ds_train.get_dataset(name)
                        name = next(names)
                        emb_dim = model_kwargs[name + '_emb_dim']
                        if emb_dim:
                            model_kwargs[name + '_num'] = \
                                len(ds_.transform_dict)
                            model_kwargs[name + '_pad_idx'] = ds_.pad
            model = model_class(*model_args, **model_kwargs)
            if device:
                model.to(device)

        # 4. Train
        need_ds = False
        seeds = [random.randrange(1, 2**32) if seed else None
                     for _ in range(len(stages))]
        for idx, (stage, seed) in enumerate(zip(stages, seeds), start=1):
            if stage:
                stage_method = stage_methods[stage - 1]
                save_to, save_to2 = \
                    (save_as + f'_{idx}(stage{stage})', save_as) \
                        if save_stages else \
                    (save_as, None)
                if need_ds:
                    ds_train = ds_test = None
                    gc.collect()
                    #torch.cuda.empty_cache()
                    ds_train, ds_test = stage_ds(idx)
                res, change_load_from = stage_method(
                   load_from, save_to, res, save_to2=save_to2, seed=seed
                )
                need_ds = stage == 3 and res and res['best_epoch'] is not None
                if change_load_from:
                    load_from = save_to

        if log_file:
            print(f'\n=== {header} TAGGER TRAINING HAS FINISHED === '
                  f'Total time: {junky.seconds_to_strtime(time.time() - start_time)} ===\n',
                  file=log_file)
            print(f"Use the `.load('{save_as}')` method to start working "
                  f'with the {header} tagger.', file=log_file)

        return res


def conll18_ud_eval(gold_file, system_file, verbose=True, counts=False):
    """A wrapper for the official
    [CoNLL18 UD Shared Task](https://universaldependencies.org/conll18/results.html)
    evaluation script.

    Args:

    **gold_file**: Name of the CoNLL-U file with the gold data.

    **system_file**: Name of the CoNLL-U file with the predicted.

    **verbose** (`bool`): Print all metrics.

    **counts** (`bool`): Print raw counts of correct/gold/system/aligned words
    instead of prec/rec/F1 for all metrics.

    If `verbose=False`, only the official CoNLL18 UD Shared Task evaluation
    metrics are printed.

    If `verbose=True` (default), more metrics are printed (as precision,
    recall, F1 score, and in case the metric is computed on aligned words
    also accuracy on these):
    - Tokens: how well do the gold tokens match system tokens.
    - Sentences: how well do the gold sentences match system sentences.
    - Words: how well can the gold words be aligned to system words.
    - UPOS: using aligned words, how well does UPOS match.
    - XPOS: using aligned words, how well does XPOS match.
    - UFeats: using aligned words, how well does universal FEATS match.
    - AllTags: using aligned words, how well does UPOS+XPOS+FEATS match.
    - Lemmas: using aligned words, how well does LEMMA match.
    - UAS: using aligned words, how well does HEAD match.
    - LAS: using aligned words, how well does HEAD+DEPREL(ignoring subtypes)
    match.
    - CLAS: using aligned words with content DEPREL, how well does
    HEAD+DEPREL(ignoring subtypes) match.
    - MLAS: using aligned words with content DEPREL, how well does
    HEAD+DEPREL(ignoring subtypes)+UPOS+UFEATS+FunctionalChildren(DEPREL+UPOS+UFEATS)
    match.
    - BLEX: using aligned words with content DEPREL, how well does
    HEAD+DEPREL(ignoring subtypes)+LEMMAS match.

    If `count=True`, raw counts of correct/gold_total/system_total/aligned
    words are printed instead of precision/recall/F1/AlignedAccuracy for all
    metrics."""
    argv = sys.argv
    sys.argv[1:] = [gold_file, system_file]
    if verbose:
        sys.argv.append('-v')
    if counts:
        sys.argv.append('-c')
    _conll18_ud_eval()
    sys.argv = argv



# -*- coding: utf-8 -*-
# MorDL project: Base model
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a base class for MorDL models.
"""
import json
from mordl.defaults import CONFIG_ATTR, LOG_FILE
import torch
import torch.nn as nn


class BaseModel(nn.Module):
    """
    The base class for MorDL models.

    Args:

    **\*args**: any mordl model args.

    **\*\*kwargs**: any mordl model keyword args.
    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        setattr(self, CONFIG_ATTR, (args, kwargs))

    def save_config(self, f, log_file=LOG_FILE):
        """Saves config in the specified file.

        Args:

        **f** (`str` | `file`): the file where the config should be saved.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        config = list(getattr(self, CONFIG_ATTR, []))
        device = next(self.parameters()).device
        if device:
            config.insert(0, str(device))
        need_close = False
        if isinstance(f, str):
            f = open(f, 'wt', encoding='utf-8')
            need_close = True
        try:
            print(json.dumps(config, sort_keys=True, indent=4), file=f)
        finally:
            if need_close:
                f.close()
        if log_file:
            print('Config saved', file=log_file)

    @classmethod
    def create_from_config(cls, f, state_dict_f=None, device=None,
                           log_file=LOG_FILE):
        """Creates model with parameters and a state dictionary previouly
        saved in the specified files.

        Args:

        **f** (`str` | `file`): the config file for the model creation.

        **state_dict_f** (`str` | `file`; default is `None`): the state
        dictionary file of the previously saved PyTorch model. If `None`, the
        model is initialized by default.

        **device** (`str`; default is `None`): the device where the model
        should be loaded after it's created, e.g. `'cuda:2'`. If `None`, the
        model is loaded to CPU.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        need_close = False
        if isinstance(f, str):
            f = open(f, 'rt', encoding='utf-8')
            need_close = True
        try:
            config = json.loads(f.read())
        finally:
            if need_close:
                f.close()
        args, kwargs = [], {}
        for cfg in config:
            if isinstance(cfg, str) and not device:
                device = cfg
            elif isinstance(cfg, list) and not args:
                args = cfg
            elif isinstance(cfg, dict) and not kwargs:
                kwargs = cfg
        if log_file:
            print('Creating model...', end=' ', file=log_file)
            log_file.flush()
        '''### backward compatibility ###########################################
        for old_kwarg, new_kwarg in zip(
            ['bn1', 'bn2', 'bn3', 'do1', 'do2', 'do3', 'emb_out_dim'],
            ['emb_bn', 'pre_bn', 'post_bn', 'emb_do', 'pre_do', 'post_do',
             'final_emb_dim']
        ):
            if old_kwarg in kwargs:
                assert new_kwarg not in kwargs, \
                   f'ERROR: The parameter {old_kwarg} in the config file ' \
                   f"can't be used together with {new_kwarg}."
            kwargs[new_kwarg] = kwargs[old_kwarg]
            del kwargs[old_kwarg]
        if 'lstm_hidden_dim' in kwargs:
            assert kwargs['lstm_hidden_dim'] \
                == kwargs['final_emb_dim'] // 2, \
               f'ERROR: Models with `lstm_hidden_dim` not equals to ' \
               f'`emb_out_dim // 2` are not supported anymore.'
            del kwargs['lstm_hidden_dim']
        ######################################################################'''
        model = cls(*args, **kwargs)
        if device:
            model.to(device)
        if log_file:
            print('done.', file=log_file)
        if state_dict_f:
            model.load_state_dict(state_dict_f, log_file=log_file)
        return model

    def save_state_dict(self, f, log_file=LOG_FILE):
        """Saves the PyTorch model's state dictionary to a file for further use
        in the inference.

        Args:

        **f** (`str` | `file`): the file where the state dictionary should be
        saved.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        if log_file:
            print('Saving state_dict...', end=' ', file=log_file)
            log_file.flush()
        torch.save(self.state_dict(), f, pickle_protocol=2)
        if log_file:
            print('done.', file=log_file)

    def load_state_dict(self, f, log_file=LOG_FILE):
        """Loads the previously saved PyTorch model's state dictionary for
        the inference.

        Args:

        **f** (`str` | `file`): the file from where the state dictionary
        should be loaded.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        if log_file:
            print('Loading state_dict...', end=' ', file=log_file)
            log_file.flush()
        device = next(self.parameters()).device
        super().load_state_dict(torch.load(f, map_location=device))
        self.eval()
        if log_file:
            print('done.', file=log_file)

    def save(self, f, log_file=LOG_FILE):
        """Saves the full model configuration to the single file.

        Args:

        **f** (`str` | `file`): the file where the model should be saved.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        if log_file:
            print('Saving model...', end=' ', file=log_file)
            log_file.flush()
        torch.save(self, f, pickle_protocol=2)
        if log_file:
            print('done.', file=log_file)

    @staticmethod
    def load(f, device=None, log_file=LOG_FILE):
        """Loads previously saved model configuration.

        Args:

        **f** (`str` | `file`): the file from where the model should be
        loaded.

        **device** (`str`; default is `None`): the device where the model
        should be loaded, e.g. `'cuda:2'`. If `None`, the model is loaded to
        CPU.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.
        """
        if log_file:
            print('Loading model...', end=' ', file=log_file)
            log_file.flush()
        model = torch.load(f, map_location=device)
        model.eval()
        if log_file:
            print('done.', file=log_file)
        return model

    def adjust_model_for_train(self, lr=.0001, betas=(0.9, 0.999), eps=1e-8,
                               weight_decay=0, amsgrad=False,
                               labels_pad_idx=-100):
        """Creates model, criterion, optimizer and scheduler for training.
        Adam optimizer is used to train the model.

        Args:

        **lr**, **betas**, **eps**, **weight_decay**, **amsgrad**: the
        hyperparams for the torch.optim.Adam optimizer.

        **labels_pad_idx** (`int`; default=-100): the index of padding element
        in the label vocabulary.
        """
        criterion = nn.CrossEntropyLoss(ignore_index=labels_pad_idx)
        optimizer = torch.optim.Adam(
            self.parameters(), lr=lr, betas=betas, eps=eps,
            weight_decay=weight_decay, amsgrad=amsgrad
        )
        scheduler = None
        return criterion, optimizer, scheduler

    def adjust_model_for_tune(self, lr=.001, momentum=.9, weight_decay=0,
                              dampening=0, nesterov=False,
                              labels_pad_idx=-100):
        """Ajusts model for post-train finetuning. Optimizer is changed to
        SGD to finetune the model.

        Args:

        **lr**, **momentum**, **weight_decay**, **dampening**, **nesterov**:
        the hyperparams for the torch.optim.SGD optimizer.

        **labels_pad_idx** (`int`; default=-100): the index of padding element
        in the label vocabulary.
        """
        criterion = nn.CrossEntropyLoss(ignore_index=labels_pad_idx)
        optimizer = torch.optim.SGD(
            self.parameters(), lr=lr, momentum=momentum,
            weight_decay=weight_decay, dampening=dampening, nesterov=nesterov
        )
        scheduler = None
        return criterion, optimizer, scheduler



# -*- coding: utf-8 -*-
# MorDL project: MISC:NE tagger
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a Named-entity tagger class.
"""
from mordl import FeatTagger


class NeTagger(FeatTagger):
    """
    The named-entity tagger class. We use the feature 'NE' of MISC field as
    the place where Named-entity tags are stored.

    Args:

    **feats_prune_coef** (`int`): feature prunning coefficient which allows to
    eliminate all features that have a low frequency. For each UPOS tag, we
    get a number of occurences of the most frequent feature from FEATS field,
    divide it by **feats_prune_coef** and use only those features, number of
    occurences of which is greater than that value, to improve the prediction
    quality.
    * `feats_prune_coef=0` means "do not use feats";
    * `feats_prune_coef=None` means "use all feats";
    * default `feats_prune_coef=6`.

    **embs**: `dict` with paths to the embeddings file as keys and
    corresponding embeddings models as values. If tagger needs to load any
    embeddings model, firstly, model is looked up it in that `dict`.

    During init, **embs** is copied to the `embs` attribute of the creating
    object, and this attribute may be used further to share already loaded
    embeddings with another taggers.
    """
    def __init__(self, feats_prune_coef=6, embs=None):
        super().__init__('MISC:NE', feats_prune_coef=feats_prune_coef,
                         embs=None)



# -*- coding: utf-8 -*-
# MorDL project: utils
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a bunch of utilities to use with PyTorch.
"""
import sys


BATCH_SIZE = 64
TRAIN_BATCH_SIZE = 32
NONE_TAG = '_'
CONFIG_ATTR = '_mordl_config'
CONFIG_EXT = '.config.json'
LOG_FILE = sys.stdout
CDICT_COEF_THRESH = .99



# -*- coding: utf-8 -*-
# MorDL project
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
MorDL provides tools for complete morphological sentence parsing and
named-entity recognition.
"""
from mordl._version import __version__
from mordl.word_embeddings import WordEmbeddings
from mordl.upos_tagger import UposTagger
from mordl.feat_tagger import FeatTagger
from mordl.feats_tagger import FeatsJointTagger as FeatsTagger
from mordl.lemma_tagger import LemmaTagger
from mordl.deprel_tagger import DeprelTagger, DeprelSeqTagger
from mordl.ne_tagger import NeTagger
from mordl.base_tagger import conll18_ud_eval

save_conllu = UposTagger.save_conllu
load_conllu = UposTagger.load_conllu
load_word_embeddings = WordEmbeddings.load
remove_rare_feats = UposTagger.remove_rare_feats



# -*- coding: utf-8 -*-
# MorDL project: Base tagger model
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a base model for MorDL taggers.
"""
from collections.abc import Iterable
from junky import CharEmbeddingRNN, CharEmbeddingCNN, get_func_params
from mordl.base_model import BaseModel
from mordl.defaults import CONFIG_ATTR
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence


class BaseTaggerSequenceModel(BaseModel):
    """
    The base model for MorDL taggers.

    Args:

    **labels_num** (`int`): number of target labels.

    **vec_emb_dim** (`int`): word-level embedding vector space dimensionality.
    If `None`, the layer is skipped.

    **alphabet_size** (`int`): length of character vocabulary. Relevant with
    not `None` **rnn_emb_dim** or **cnn_emb_dim**.

    **char_pad_idx** (`int`): index of padding element in the character
    vocabulary. Relevant with not `None` **rnn_emb_dim** or **cnn_emb_dim**.

    **rnn_emb_dim** (`int`): character RNN (LSTM) embedding dimensionality. If
    `None`, the layer is skipped.

    **cnn_emb_dim** (`int`): character CNN embedding dimensionality. If
    `None`, the layer is skipped.

    **cnn_kernels** (`list([int])`): CNN kernel sizes. By default,
    `cnn_kernels=[1, 2, 3, 4, 5, 6]`. Relevant with not `None`
    **cnn_emb_dim**.

    **tag_emb_params** (`dict({'dim': int, 'num': int, 'pad_idx': int})` |
    `list([dict])`): params of the embedding layers for additional
    `junky.dataset.TokenDataset` outputs. If `None`, the layers are not
    created.

    **emb_out_dim** (`int`): output embedding dimensionality. Default
    `emb_out_dim=512`.

    **lstm_hidden_dim** (`int`): Bidirectional LSTM hidden size. Default
    `lstm_hidden_dim=256`.

    **lstm_layers** (`int`): number of Bidirectional LSTM layers. Default
    `lstm_layers=1`.

    **lstm_do** (`float`): dropout between LSTM layers. Only relevant, if
    `lstm_layers` > `1`.

    **bn1** (`bool`): whether batch normalization layer should be applied
    after the embedding layer. Default `bn1=True`.

    **do1** (`float`): dropout rate after the first batch normalization
    layer `bn1`. Default `do1=.2`.

    **bn2** (`bool`): whether batch normalization layer should be applied
    after the linear layer before LSTM layer. Default `bn2=True`.

    **do2** (`float`): dropout rate after the second batch normalization
    layer `bn2`. Default `do2=.5`.

    **bn3** (`bool`): whether batch normalization layer should be applied
    after the LSTM layer. Default `bn3=True`.

    **do3** (`float`): dropout rate after the third batch normalization
    layer `bn3`. Default `do3=.4`.
    """
    def __init__(self, labels_num, vec_emb_dim=None,
                 alphabet_size=0, char_pad_idx=0, rnn_emb_dim=None,
                 cnn_emb_dim=None, cnn_kernels=[1, 2, 3, 4, 5, 6],
                 tag_emb_params=None, emb_out_dim=512,
                 lstm_bidirectional=True, lstm_hidden_dim=256, lstm_layers=1,
                 lstm_do=0, bn1=True, do1=.2, bn2=True, do2=.5,
                 bn3=True, do3=.4):
        if isinstance(cnn_kernels, Iterable):
            cnn_kernels = list(cnn_kernels)
        args, kwargs = \
            get_func_params(BaseTaggerSequenceModel.__init__, locals())
        super().__init__(*args, **kwargs)

        self.vec_emb_dim = vec_emb_dim

        if rnn_emb_dim:
            self._rnn_emb_l = \
                CharEmbeddingRNN(alphabet_size=alphabet_size,
                                 emb_dim=rnn_emb_dim,
                                 pad_idx=char_pad_idx)
        else:
            self._rnn_emb_l = None
            rnn_emb_dim = 0

        if cnn_emb_dim:
            self._cnn_emb_l = \
                CharEmbeddingCNN(alphabet_size=alphabet_size,
                                 emb_dim=cnn_emb_dim,
                                 pad_idx=char_pad_idx,
                                 kernels=cnn_kernels)
        else:
            self._cnn_emb_l = None
            cnn_emb_dim = 0

        self._tag_emb_l = self._tag_emb_ls = None
        tag_emb_dim = 0
        if tag_emb_params:
            if isinstance(tag_emb_params, dict):
                tag_emb_dim = tag_emb_params['dim'] or 0
                if tag_emb_dim:
                    self._tag_emb_l = \
                        nn.Embedding(tag_emb_params['num'], tag_emb_dim,
                                     padding_idx=tag_emb_params['pad_idx'])
            else:
                self._tag_emb_ls = nn.ModuleList()
                for emb_params in tag_emb_params:
                    tag_emb_dim_ = emb_params['dim']
                    if tag_emb_dim_:
                        tag_emb_dim += tag_emb_dim_
                        self._tag_emb_ls.append(
                            nn.Embedding(emb_params['num'], tag_emb_dim_,
                                         padding_idx=emb_params['pad_idx'])
                        )
                    else:
                        self._tag_emb_ls.append(None)

        self._bn1 = \
            nn.BatchNorm1d(num_features=vec_emb_dim
                                      + rnn_emb_dim
                                      + cnn_emb_dim
                                      + tag_emb_dim) if bn1 else None
        self._do1 = nn.Dropout(p=do1) if do1 else None

        self._emb_fc_l = nn.Linear(
            in_features=vec_emb_dim + rnn_emb_dim + cnn_emb_dim
                                    + tag_emb_dim,
            out_features=emb_out_dim
        )
        self._bn2 = \
            nn.BatchNorm1d(num_features=emb_out_dim) if bn2 else None
        self._do2 = nn.Dropout(p=do2) if do2 else None

        self._lstm_l = nn.LSTM(input_size=emb_out_dim,
                               hidden_size=lstm_hidden_dim,
                               num_layers=lstm_layers, batch_first=True,
                               dropout=lstm_do,
                               bidirectional=lstm_bidirectional)
        if lstm_bidirectional:
            lstm_hidden_dim *= 2
        self._T = nn.Linear(emb_out_dim, lstm_hidden_dim)
        nn.init.constant_(self._T.bias, -1)

        self._bn3 = \
            nn.BatchNorm1d(num_features=lstm_hidden_dim) if bn3 else None
        self._do3 = nn.Dropout(p=do3) if do3 else None

        self._out_l = nn.Linear(in_features=lstm_hidden_dim,
                                out_features=labels_num)

        setattr(self, CONFIG_ATTR, (args, kwargs))

    def forward(self, x, x_lens, x_ch, x_ch_lens, *x_t):
        """
        x:    [batch[seq[w_idx + pad]]]
        lens: [seq_word_cnt]
        x_ch: [batch[seq[word[ch_idx + pad] + word[pad]]]]
        x_ch_lens: [seq[word_char_count]]
        *x_t:  [batch[seq[upos_idx]]], ...
        """
        device = next(self.parameters()).device

        x_ = []
        if self.vec_emb_dim:
            assert x.shape[2] == self.vec_emb_dim, \
                   'ERROR: Invalid vector size: {} whereas vec_emb_dim = {}' \
                       .format(x.shape[2], self.vec_emb_dim)
            x_.append(x)
        if self._rnn_emb_l:
            x_.append(self._rnn_emb_l(x_ch, x_ch_lens))
        if self._cnn_emb_l:
            x_.append(self._cnn_emb_l(x_ch, x_ch_lens))
        if self._tag_emb_l:
            x_.append(self._tag_emb_l(x_t[0]))
        elif self._tag_emb_ls:
            for l_, x_t_ in zip(self._tag_emb_ls, x_t):
                if l_:
                    x_.append(l_(x_t_))

        x = x_[0] if len(x_) == 1 else torch.cat(x_, dim=-1)

        if self._bn1:
            x = x.transpose(1, 2)  # (N, L, C) to (N, C, L)
            x = self._bn1(x)
            x = x.transpose(1, 2)  # (N, C, L) to (N, L, C)
        if self._do1:
            x = self._do1(x)

        x = self._emb_fc_l(x)
        if self._bn2:
            x = x.transpose(1, 2)  # (N, L, C) to (N, C, L)
            x = self._bn2(x)
            x = x.transpose(1, 2)  # (N, C, L) to (N, L, C)
        x = F.relu(x)
        if self._do2:
            x = self._do2(x)

        x_ = pack_padded_sequence(x, x_lens, batch_first=True,
                                  enforce_sorted=False)
        output, (h_n, c_n) = self._lstm_l(x_)
        # h_n.shape => [batch size, num layers * num directions, hidden size]

        ## 1. if we want to use h_n:
        x_ = torch.cat((h_n[-2, :, :], h_n[-1, :, :]), dim=1) \
                 if self._lstm_l.bidirectional else \
             h_n[-1, :, :]
        # x_.shape => [batch size, hidden size]
        #####

        '''
        ## 2. if we want to use output:
        x_, _ = pad_packed_sequence(output, batch_first=True)
        x_ = x_[:, -1, :]
        # or
        x_ = torch.max(x_, dim=1)[0]
        # or
        x_ = torch.mean(x_, dim=1)
        # or
        x_ = torch.sum(x_, dim=1)
        #####
        '''

        gate = torch.sigmoid(self._T(x))
        x = x_ * gate + x * (1 - gate)

        if self._bn3:
            x = self._bn3(x)
        if self._do3:
            x = self._do3(x)

        x = self._out_l(x)

        return x



# -*- coding: utf-8 -*-
# MorDL project: DEPREL tagger
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides classes of DEPREL taggers.
"""
from copy import deepcopy
import itertools
import junky
from junky import get_func_params
from mordl import FeatTagger
from mordl.defaults import BATCH_SIZE, LOG_FILE, TRAIN_BATCH_SIZE
from mordl.deprel_tagger_model import DeprelTaggerModel
import time
import torch
from typing import Iterator


# LAS on *SynTagRus*: 97.60
# with supp_tagger of `DeprelSeqTagger` type: 97.60
class DeprelTagger(FeatTagger):
    """
    The DEPREL tagger class.

    Args:

    **feats_prune_coef** (`int`; default is `6`): the feature prunning
    coefficient which allows to eliminate all features that have a low
    frequency. To improve the prediction quality, we get a number of
    occurences of the most frequent feature from the FEATS field for each UPOS
    tag, divide that number by **feats_prune_coef**, and use only those
    features, the number of occurences of which is greater than that value.
    * `feats_prune_coef=0` means "do not use feats";
    * `feats_prune_coef=None` means "use all feats";
    * default `feats_prune_coef=6`.

    **supp_tagger** (default is `None`): an object of another DEPREL tagger
    which method `.predict()` has the same signature as
    `DeprelTagger.predict()` and no excess `'root'` tags in the return. The
    instance of `DeprelSeqTagger` class may be used here.

    **embs** (`dict({str: object}); default is `None`): the `dict` with paths
    to embeddings files as keys and corresponding embedding models as values.
    If the tagger needs to load any embedding model, firstly, the model is
    looked up it in that `dict`.

    During init, **embs** is copied to the `embs` attribute of the creating
    object, and this attribute may be used further to share already loaded
    embeddings with another taggers.
    """
    def __init__(self, feats_prune_coef=6, supp_tagger=None, embs=None):
        super().__init__('DEPREL', feats_prune_coef=feats_prune_coef,
                         embs=embs)
        self._supp_tagger = supp_tagger

    def predict(self, corpus, use_cdict_coef=False, with_orig=False,
                batch_size=BATCH_SIZE, split=None, clone_ds=False,
                save_to=None, log_file=LOG_FILE):
        """Predicts tags in the DEPREL field of the corpus.

        Args:

        **corpus**: the corpus which will be used for the feature extraction
        and predictions. May be either the name of the file in *CoNLL-U*
        format or the `list`/`iterator` of sentences in *Parsed CoNLL-U*.

        **use_cdict_coef** (`bool` | `float`; default is `False`): if `False`,
        we use our prediction only. If `True`, we replace our prediction to
        the value returned by the `corpuscula.CorpusDict.predict_<field>()`
        method if its `coef` >= `.99`. Also, you can specify your own
        threshold as the value of the param.

        **with_orig** (`bool`; default is `False`): if `True`, instead of just
        the sequence with predicted labels, return the sequence of tuples
        where the first element is the sentence with predicted labels and the
        second element is the original sentence. **with_orig** can be `True`
        only if **save_to** is `None`.

        **batch_size** (`int`; default is `64`): the number of sentences per
        batch.

        **split** (`int`; default is `None`): the number of lines in sentences
        split. Allows to process a large dataset in pieces ("splits"). If
        **split** is `None` (default), all the dataset is processed without
        splits.

        **clone_ds** (`bool`; default is `False`): if `True`, the dataset is
        cloned and transformed. If `False`, `transform_collate` is used
        without cloning the dataset. There is no big differences between the
        variants. Both should produce identical results.

        **save_to** (`str`; default is `None`): the file name where the
        predictions will be saved.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        Returns the corpus with tags predicted in the DEPREL field.
        """
        assert self._ds is not None, \
               "ERROR: The tagger doesn't have a dataset. Call the train() " \
               'method first'
        assert not with_orig or save_to is None, \
               'ERROR: `with_orig` can be True only if save_to is None'
        args, kwargs = get_func_params(FeatTagger.predict, locals())

        kwargs['save_to'] = None

        corpus2 = None
        if self._supp_tagger:
            kwargs2 = kwargs.copy()
            kwargs2['with_orig'] = True
            corpus2 = self._supp_tagger.predict(*args, **kwargs2)

        corpus = super().predict(*args, **kwargs)

        def add_root(corpus):
            for sent in corpus:
                sent0 = sent[0] if isinstance(sent, tuple) \
                               and not isinstance(sent[0], tuple) \
                               and not isinstance(sent[1], tuple) else \
                        sent
                if isinstance(sent0, tuple):
                    sent0 = sent0[0]
                if corpus2:
                    sent2 = next(corpus2)[1]
                    if isinstance(sent2, tuple):
                        sent2 = sent2[0]
                for i, tok in enumerate(sent0):
                    if tok['HEAD'] == '0':
                        tok['DEPREL'] = 'root'
                    elif corpus2 and tok['DEPREL'] == 'root':
                        tok['DEPREL'] = sent2[i]['DEPREL']
                yield sent

        corpus = add_root(corpus)

        if save_to:
            self.save_conllu(corpus, save_to, log_file=None)
            corpus = self._get_corpus(save_to, asis=True, log_file=log_file)

        return corpus

NODES_UP, NODES_DOWN = 2, 2
PAD = '[PAD]'
PAD_TOKEN = {'ID': '0', 'FORM': PAD, 'LEMMA': PAD,
                         'UPOS': PAD, 'FEATS': {}, 'DEPREL': None}

class DeprelSeqTagger(FeatTagger):
    """
    A DEPREL tagger class with sequence model.

    **WARNING**: Works long, takes huge amoont of RAM, but accuracy is 
    less than of `DeprelTagger`. But may be useful as suppplement tagger
    for it.

    Args:

    **feats_prune_coef** (`int`): feature prunning coefficient which allows to
    eliminate all features that have a low frequency. For each UPOS tag, we
    get a number of occurences of the most frequent feature from FEATS field,
    divide it by **feats_prune_coef** and use only those features, number of
    occurences of which is greater than that value, to improve the prediction
    quality.
    * `feats_prune_coef=0` means "do not use feats";
    * `feats_prune_coef=None` means "use all feats";
    * default `feats_prune_coef=6`.
    """
    def __init__(self, feats_prune_coef=6):
        super().__init__('DEPREL', feats_prune_coef=feats_prune_coef)
        self._model_class = DeprelTaggerModel

    @staticmethod
    def _preprocess_corpus(corpus):

        def nodes_down(sent, id_, ids, nodes, span_len):
            link_ids = nodes.get(id_, [])
            res = []
            for link_id in link_ids:
                idx = ids[link_id]
                token = sent[idx]
                if span_len == 1:
                    res.append([token])
                else:
                    for nodes_down_ in nodes_down(sent, link_id, ids,
                                                  nodes, span_len - 1):
                        res.append([token] + nodes_down_)
            if not res:
                res = [[PAD_TOKEN] * span_len]
            return res

        def next_sent(sent, upper_sent, id_, ids, nodes):
            link_ids = nodes.get(id_, [])
            for link_id in link_ids:
                idx = ids[link_id]
                token = sent[idx]
                label = token['DEPREL']
                s = upper_sent[-NODES_UP:]
                s.append(token)
                for nodes_down_ in nodes_down(sent, link_id, ids,
                                              nodes, NODES_DOWN):
                    s_ = ([PAD_TOKEN] * (NODES_UP + 1 - len(s))) \
                       + s[:] + nodes_down_
                    yield s_, idx, label
                for data_ in next_sent(sent, s, link_id, ids, nodes):
                    yield data_

        res_corpus, labels, restore_data = [], [], []
        for i, sent in enumerate(corpus):
            if isinstance(sent, tuple):
                sent = sent[0]
            root_id, root_token, ids, nodes = None, None, {}, {}
            for idx, token in enumerate(sent):
                id_, head = token['ID'], token['HEAD']
                if head:
                    if not token['FORM']:
                        token['FORM'] = PAD
                    if head == '0':
                        root_id, root_token = id_, token
                    ids[id_] = idx
                    nodes.setdefault(head, []).append(id_)
            if root_id:
                for s, idx, label in next_sent(sent, [root_token],
                                               root_id, ids, nodes):
                    res_corpus.append(s)
                    labels.append(label)
                    restore_data.append((i, idx))

        return res_corpus, labels, restore_data

    @staticmethod
    def _postprocess_corpus(corpus, labels, restore_data):
        for label, (i, idx) in zip(labels, restore_data):
            token = corpus[i][idx]
            if not isinstance(token['DEPREL'], list):
                token['DEPREL'] = [label]
            else:
                token['DEPREL'].append(label)
        for sent in corpus:
            if isinstance(sent, tuple):
                sent = sent[0]
            for tok in sent:
                if tok['HEAD'] == '0':
                    tok['DEPREL'] = 'root'
                else:
                    deprel = tok['DEPREL']
                    if isinstance(deprel, list):
                        tok['DEPREL'] = max(set(deprel), key=deprel.count)
        return corpus

    def _prepare_corpus(self, corpus, fields, tags_to_remove=None):
        res = super()._prepare_corpus(corpus, fields,
                                      tags_to_remove=tags_to_remove)
        res = list(res)
        res[-1] = [x[NODES_UP] for x in res[-1]]
        return tuple(res)

    def load(self, name, device=None,
             dataset_emb_path=None, dataset_device=None, log_file=LOG_FILE):
        """Loads tagger's internal state saved by its `.save()` method.

        Args:

        **name** (`str`): name of the previously saved internal state.

        **device**: a device for the loaded model if you want to override
        the value from config.

        **dataset_emb_path**: a path where dataset's embeddings to load from
        if you want to override the value from config.

        **dataset_device**: a device for the loaded dataset if you want to
        override the value from config.

        **log_file**: a stream for info messages. Default is `sys.stdout`.
        """
        args, kwargs = get_func_params(DeprelTagger.load, locals())
        super(self.__class__.__base__, self).load(DeprelTaggerModel,
                                                  *args, **kwargs)

    def predict(self, corpus, with_orig=False, batch_size=BATCH_SIZE,
                split=None, clone_ds=False, save_to=None, log_file=LOG_FILE):
        """Predicts tags in the DEPREL field of the corpus.

        Args:

        **corpus**: a corpus which will be used for feature extraction and
        predictions. May be either a name of the file in *CoNLL-U* format or a
        list/iterator of sentences in *Parsed CoNLL-U*.

        **with_orig** (`bool`): if `True`, instead of only a sequence with
        predicted labels, returns a sequence of tuples where the first element
        is a sentence with predicted labels and the second element is the
        original sentence. `with_orig` can be `True` only if `save_to` is
        `None`. Default `with_orig=False`.

        **batch_size** (`int`): number of sentences per batch. Default
        `batch_size=64`.

        **split** (`int`): number of lines in each split. Allows to process a
        large dataset in pieces ("splits"). Default `split=None`, i.e. process
        full dataset without splits.

        **clone_ds** (`bool`): if `True`, the dataset is cloned and
        transformed. If `False`, `transform_collate` is used without cloning
        the dataset. There is no big differences between the variants. Both
        should produce identical results.

        **save_to**: file name where the predictions will be saved.

        **log_file**: a stream for info messages. Default is `sys.stdout`.

        Returns corpus with tags predicted in the DEPREL field.
        """
        assert self._ds is not None, \
               "ERROR: The tagger doesn't have a dataset. Call the train() " \
               'method first'
        assert self._model, \
               "ERROR: The tagger doesn't have a model. Call the train() " \
               'method first'
        assert not with_orig or save_to is None, \
               'ERROR: `with_orig` can be True only if save_to is None'
        args, kwargs = get_func_params(DeprelTagger.predict, locals())

        if self._feats_prune_coef != 0:
            key_vals = self._ds.get_dataset('t_0').transform_dict
            corpus = self._get_corpus(corpus, asis=True, log_file=log_file)
            corpus = self._transform_upos(corpus, key_vals=key_vals)

        field = 'DEPREL'
        add_fields = self._normalize_field_names('UPOS')

        def process(corpus):

            corpus = self._get_corpus(corpus, asis=True, log_file=log_file)
            device = next(self._model.parameters()).device or junky.CPU

            ds_y = self._ds.get_dataset('y')
            if clone_ds:
                ds = self._ds.clone()
                ds.remove('y')

            for start in itertools.count(step=split if split else 1):
                if isinstance(corpus, Iterator):
                    if split:
                        corpus_ = []
                        for i, sentence in enumerate(corpus, start=1):
                            corpus_.append(sentence)
                            if i == split:
                                break
                    else:
                        corpus_ = list(corpus)
                else:
                    if split:
                        corpus_ = corpus[start:start + split]
                    else:
                        corpus_ = corpus
                if not corpus_:
                    break

                orig_corpus = corpus_
                corpus_, _, restore_data = self._preprocess_corpus(corpus_)
                
                res = \
                    junky.extract_conllu_fields(
                        corpus_, fields=add_fields,
                        with_empty=True, return_nones=True
                    )
                sentences, tags, empties, nones = \
                    res[0], res[1:-2], res[-2], res[-1]
                if clone_ds:
                    self._transform(
                        sentences, tags=tags, batch_size=batch_size, ds=ds,
                        log_file=log_file
                    )
                    loader = ds.create_loader(batch_size=batch_size,
                                              shuffle=False)
                else:
                    loader = self._transform_collate(
                        sentences, tags=tags, batch_size=batch_size,
                        log_file=log_file
                    )
                preds = []
                for batch in loader:
                    batch = junky.to_device(batch, device)
                    with torch.no_grad():
                        pred = self._model(*batch)
                    pred_indices = pred.argmax(-1)
                    preds.extend(pred_indices.cpu().numpy().tolist())
                values = ds_y.reconstruct(preds)
                if with_orig:
                    res_corpus_ = deepcopy(orig_corpus)
                    res_corpus_ = \
                        self._postprocess_corpus(res_corpus_,
                                                 values, restore_data)
                    for orig_sentence, sentence in zip(orig_corpus,
                                                       res_corpus_):
                        yield sentence, orig_sentence
                else:
                    res_corpus_ = \
                        self._postprocess_corpus(orig_corpus,
                                                 values, restore_data)
                    for sentence in res_corpus_:
                        yield sentence

        corpus = process(corpus)

        if self._feats_prune_coef != 0:
            corpus = self._restore_upos(corpus)

        if save_to:
            self.save_conllu(corpus, save_to, log_file=None)
            corpus = self._get_corpus(save_to, asis=True, log_file=log_file)
        return corpus

    def train(self, save_as,
              device=None, control_metric='accuracy', max_epochs=None,
              min_epochs=0, bad_epochs=5, batch_size=TRAIN_BATCH_SIZE,
              max_grad_norm=None, tags_to_remove=None, word_emb_type='bert',
              word_emb_path='xlm-roberta-base', word_transform_kwargs=None,
                  # BertDataset.transform() (for BERT-descendant models)
                  # params:
                  # {'max_len': 0, 'batch_size': 64, 'hidden_ids': '10',
                  #  'aggregate_hiddens_op': 'cat',
                  #  'aggregate_subtokens_op': 'absmax', 'to': junky.CPU,
                  #  'loglevel': 1}
                  # WordDataset.transform() (for other models) params:
                  # {'check_lower': True}
              stage1_params=None,
                  # {'lr': .0001, 'betas': (0.9, 0.999), 'eps': 1e-8,
                  #  'weight_decay': 0, 'amsgrad': False,
                  #  'max_epochs': None, 'min_epochs': None,
                  #  'bad_epochs': None, 'batch_size': None,
                  #  'max_grad_norm': None}
              stage2_params=None,
                  # {'lr': .001, 'momentum': .9, 'weight_decay': 0,
                  #  'dampening': 0, 'nesterov': False,
                  #  'max_epochs': None, 'min_epochs': None,
                  #  'bad_epochs': None, 'batch_size': None,
                  #  'max_grad_norm': None}
              stage3_params={'save_as': None},
                  # {'save_as': None, 'epochs': 3, 'batch_size': 8,
                  #  'lr': 2e-5, 'betas': (0.9, 0.999), 'eps': 1e-8,
                  #  'weight_decay': .01, 'amsgrad': False,
                  #  'num_warmup_steps': 3, 'max_grad_norm': 1.}
              stages=[1, 2, 3, 1, 2], save_stages=False, load_from=None,
              learn_on_padding=True, remove_padding_intent=False,
              seed=None, start_time=None, keep_embs=False, log_file=LOG_FILE,
              rnn_emb_dim=None, cnn_emb_dim=None, cnn_kernels=range(1, 7),
              upos_emb_dim=300, emb_bn=True, emb_do=.2,
              final_emb_dim=512, pre_bn=True, pre_do=.5,
              lstm_layers=1, lstm_do=0, tran_layers=0, tran_heads=8,
              post_bn=True, post_do=.4):
        """Creates and trains the DEPREL tagger model.

        During training, the best model is saved after each successful epoch.

        *Training's args*:

        **save_as** (`str`): the name using for save the model's head. Refer
        to the `.save()` method's help for the broad definition (see the
        **name** arg there).

        **device** (`str`, default is `None`): the device for the model. E.g.:
        'cuda:0'. If `None`, we don't move the model to any device (it is
        placed right where it's created).

        **control_metric** (`str`; default is `accuracy`): the metric that
        control training. Any that is supported by the `junky.train()` method.
        In the moment, it is: 'accuracy', 'f1', 'loss', 'precision', and
        'recall'.

        **max_epochs** (`int`; default is `None`): the maximal number of
        epochs for the model's head training (stages types `1` and `2`). If
        `None` (default), the training would be linger until **bad_epochs**
        has met, but no less than **min_epochs**.

        **min_epochs** (`int`; default is `0`): the minimal number of training
        epochs for the model's head training (stages types `1` and `2`).

        **bad_epochs** (`int`; default is `5`): the maximal allowed number of
        bad epochs (epochs when chosen **control_metric** is not became
        better) in a row for the model's head training (stages types `1` and
        `2`).

        **batch_size** (`int`; default is `32`): the number of sentences per
        batch for the model's head training (stages types `1` and `2`).

        **max_grad_norm** (`float`; default is `None`): the gradient clipping
        parameter for the model's head training (stages types `1` and `2`).

        **tags_to_remove** (`dict({str: str}) | dict({str: list([str])})`;
        default is `None`): the tags, tokens with those must be removed from
        the corpus. It's the `dict` with field names as keys and values you
        want to remove. Applied only to fields with atomic values (like
        *UPOS*). This argument may be used, for example, to remove some
        infrequent or just excess tags from the corpus. Note, that we remove
        the tokens from the train corpus completely, not just replace those
        tags to `None`.

        *Word embedding params*:

        **word_emb_type** (`str`; default is `'bert'`): one of (`'bert'` |
        `'glove'` | `'ft'` | `'w2v'`) embedding types.

        **word_emb_path** (`str`): the path to the word embeddings storage.

        **word_transform_kwargs** (`dict`; default is `None`): keyword
        arguments for the `.transform()` method of the dataset created for
        sentences to word embeddings conversion. See the `.transform()` method
        of either `junky.datasets.BertDataset` (if **word_emb_path** is
        `'bert'`) or `junky.datasets.WordDataset` (otherwise) if you want to
        learn allowed values for the parameter. If `None`, the `.transform()`
        method use its defaults.

        *Training stages params*:

        **stage1_param** (`dict`; default is `None`): keyword arguments for
        the `BaseModel.adjust_model_for_train()` method. If `None`, the
        defaults are used. Also, you can specify here new values for the
        arguments **max_epochs**, **min_epochs**, **bad_epochs**,
        **batch_size**, and **max_grad_norm** that will be used only on stages
        of type `1`.

        **stage2_param** (`dict`; default is `None`): keyword arguments for
        the `BaseModel.adjust_model_for_tune()` method. If `None`, the
        defaults are used. Also, you can specify here new values for the
        arguments **max_epochs**, **min_epochs**, **bad_epochs**,
        **batch_size**, and **max_grad_norm** that will be used only on stages
        of type `2`.

        **stage3_param** (`dict`; default is `None`): keyword arguments for
        the `WordEmbeddings.full_tune()` method. If `None`, the defaults are
        used.

        **stages** (`list([int]`; default is `[1, 2, 3, 1, 2]`): what stages
        we should use during training and in which order. On the stage type
        `1` the model head is trained with *Adam* optimizer; the stage type
        `2` is similar, but the optimizer is *SGD*; the stage type `3` is only
        relevant when **word_emb_type** is `'bert'` and we want to tune the
        whole model. Stage type `0` defines the skip-stage, i.e. there would
        be no real training on it. It is used when you need reproducibility
        and want to continue train the model from some particular stage. In
        this case, you specify the name of the model saved on that stage in
        the parametere **load_from**, and put zeros into the **stages** list
        on the places of already finished ones. One more time: it is used for
        reproducibility only, i.e. when you put some particular value to the
        **seed** param and want the data order in bathes be equivalent with
        data on the stages from the past trainings.

        **save_stages** (`bool`; default is `False`): if we need to keep the
        best model of each stage beside of the overall best model. The names
        of these models would have the suffix `_<idx>(stage<stage_type>)`
        where `<idx>` is an ordinal number of the stage. We can then use it to
        continue training from any particular stage number (changing next
        stages or their parameters) using the parameter **load_from**. Note
        that we save only stages of the head model. The embedding model as a
        part of the full model usually tune only once, so we don't make its
        copy.

        **load_from** (`str`; default is `None`): if you want to continue
        training from one of previously saved stages, you can specify the name
        of the model from that stage. Note, that if your model is already
        trained on stage type `3`, then you want to set param
        **word_emb_path** to `None`. Otherwise, you'll load wrong embedding
        model. Any other params of the model may be overwritten (and most
        likely, this would cause error), but they are equivalent when the
        training is just starts and when it's continues. But the
        **word_emb_path** is different if you already passed stage type `3`,
        so don't forget to set it to `None` in that case. (Example: you want
        to repeat training on stage no `5`, so you specify in the
        **load_from** param something like `'model_4(stage1)'` and set the
        **word_emb_path** to `None` and the **stages_param** to
        `'[0, 0, 0, 0, 2]'` (or, if you don't care of reproducibility, you
        could just specify `[2]` here).

        *Other options*:

        **learn_on_padding** (`bool`; default is `True`): while training, we
        can calculate loss either taking in account predictions made for
        padding tokens or without it. The common practice is don't use padding
        when calculate loss. However, we note that using padding sometimes
        makes the resulting model performance slightly better.

        **remove_padding_intent** (`bool`; default is `False`): if you set
        **learn_on_padding** param to `False`, you may want not to use padding
        intent during training at all. I.e. padding tokens would be tagged
        with some of real tags, and they would just ignored during computing
        loss. As a result, the model would have the output dimensionality of
        the final layer less by one. On the first sight, such approach could
        increase the performance, but in our experiments, such effect appeared
        not always.

        **seed** (`int`; default is `None`): init value for the random number
        generator if you need reproducibility. Note that each stage will have
        its own seed value, and the **seed** param is used to calculate these
        values.

        **start_time** (`float`; default is `None`): the result of
        `time.time()` to start with. If `None`, the arg will be init anew.

        **keep_embs** (`bool`; default is `False`): by default, after creating
        `Dataset` objects, we remove word embedding models to free memory.
        With `keep_embs=False` this operation is omitted, and you can use
        `.embs` attribute for share embedding models with other objects.

        **log_file** (`file`; default is `sys.stdout`): the stream for info
        messages.

        *The model hyperparameters*:

        **rnn_emb_dim** (`int`; default is `None`): the internal character RNN
        (LSTM) embedding dimensionality. If `None`, the layer is skipped.

        **cnn_emb_dim** (`int`; default is `None`): the internal character CNN
        embedding dimensionality. If `None`, the layer is skipped.

        **cnn_kernels** (`list([int])`; default is `[1, 2, 3, 4, 5, 6]`): CNN
        kernel sizes of the internal CNN embedding layer. Relevant if
        **cnn_emb_dim** is not `None`.

        **upos_emb_dim** (`int`; default is `300`): the auxiliary UPOS label
        embedding dimensionality.

        **emb_bn** (`bool`; default is `True`): whether batch normalization
        layer should be applied after the embedding concatenation.

        **emb_do** (`float`; default is `.2`): the dropout rate after the
        embedding concatenation.

        **final_emb_dim** (`int`; default is `512`): the output dimesionality
        of the linear transformation applying to concatenated embeddings.

        **pre_bn** (`bool`; default is `True`): whether batch normalization
        layer should be applied before the main part of the algorithm.

        **pre_do** (`float`; default is `.5`): the dropout rate before the
        main part of the algorithm.

        **lstm_layers** (`int`; default is `1`): the number of Bidirectional
        LSTM layers. If `None`, they are not created.

        **lstm_do** (`float`; default is `0`): the dropout between LSTM
        layers. Only relevant, if `lstm_layers` > `1`.

        **tran_layers** (`int`; default is `None`): the number of Transformer
        Encoder layers. If `None`, they are not created.

        **tran_heads** (`int`; default is `8`): the number of attention heads
        of Transformer Encoder layers. Only relevant, if `tran_layers` > `1`.

        **post_bn** (`bool`; default is `True`): whether batch normalization
        layer should be applied after the main part of the algorithm.

        **post_do** (`float`; default is `.4`): the dropout rate after the
        main part of the algorithm.

        The method returns the train statistics.
        """
        if not start_time:
            start_time = time.time()
        args, kwargs = get_func_params(DeprelTagger.train, locals())

        if self._train_corpus:
            self._train_corpus, _, _ = \
                self._preprocess_corpus(self._train_corpus)
        if self._test_corpus:
            self._test_corpus, _, _ = \
                self._preprocess_corpus(self._test_corpus)

        return super().train(*args, **kwargs)
